##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Stats implementations
"""
from __future__ import absolute_import

from ..stats import AbstractStats


class MySQLStats(AbstractStats):

    def get_db_size(self):
        """Returns the approximate size of the database in bytes"""
        conn, cursor = self.connmanager.open()
        try:
            cursor.execute("SHOW TABLE STATUS")
            description = [i[0] for i in cursor.description]
            # The MySQL connector C extension returns these as bytes
            # on python 3
            description = [x.decode('ascii') if not isinstance(x, str) else x
                           for x in description]
            rows = cursor.fetchall()
        finally:
            self.connmanager.close(conn, cursor)
        data_column = description.index('Data_length')
        index_column = description.index('Index_length')
        # We also got back views, which have no size.
        return sum([row[data_column] + row[index_column] for row in rows
                    if row[data_column] and row[index_column]], 0)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Stats implementations
"""
from __future__ import absolute_import

from ..stats import AbstractStats


class OracleStats(AbstractStats):

    def get_db_size(self):
        """Returns the approximate size of the database in bytes"""
        conn, cursor = self.connmanager.open(
            self.connmanager.isolation_read_only)
        try:
            stmt = """
            SELECT SUM(BYTES)
            FROM USER_SEGMENTS
            """
            cursor.execute(stmt)
            res = cursor.fetchone()[0]
            if res is None:
                res = 0
            else:
                res = int(res)
            return res
        finally:
            self.connmanager.close(conn, cursor)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from unittest import skipUnless

from ..adapter import Sqlite3Adapter as Adapter
from ..drivers import Sqlite3Driver
from ...tests import test_adapter

class MockConnection(object):

    @property
    def cursor(self):
        return self

    @property
    def connection(self):
        return self

    in_transaction = False
    in_critical_phase = True
    rolled_back = False

    def execute(self, query, _args=()):
        if query.startswith('UPDATE'):
            self.in_transaction = True

    def fetchall(self):
        return ()

    def commit(self):
        self.in_transaction = False
        self.in_critical_phase = False

    def rollback_quietly(self):
        self.rolled_back = True

class MockBlobHelper(object):

    txn_has_blobs = False

@skipUnless(Sqlite3Driver.STATIC_AVAILABLE, "Driver not available")
class TestAdapter(test_adapter.AdapterTestBase):

    def _makeOne(self, options):
        return Adapter(":memory:",
                       options=options, pragmas={})

    def _test_lock_database_and_move_ends_critical_section_on_commit(self, commit):
        from relstorage.options import Options
        options = Options()
        options.driver = 'gevent sqlite3'
        adapter = self._makeOne(None)
        store_conn = MockConnection()
        load_conn = MockConnection()
        assert store_conn.in_critical_phase

        result = adapter.lock_database_and_move(
            store_conn,
            load_conn,
            MockBlobHelper(),
            (b'username', b'desc', b'ext'),
            commit=commit
        )

        self.assertIsNotNone(result)
        self.assertEqual(store_conn.in_critical_phase, not commit)
        self.assertEqual(load_conn.rolled_back, commit)

    def test_lock_database_and_move_ends_critical_section_on_commit(self):
        self._test_lock_database_and_move_ends_critical_section_on_commit(True)

    def test_lock_database_and_move_keeps_critical_section_wo_commit(self):
        self._test_lock_database_and_move_ends_critical_section_on_commit(False)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import ZConfig.datatypes
from ZODB.DemoStorage import DemoStorage

from relstorage.options import Options
from relstorage.storage import RelStorage

try:
    from urllib import parse as urlparse
    from urllib.parse import parse_qsl
except ImportError:
    # Py2
    import urlparse
    from urlparse import parse_qsl


datatypes = ZConfig.datatypes.stock_datatypes.copy()
datatypes['string-list'] = lambda s: tuple(s.split(','))

def asBoolean(s):
    if s == '0':
        return False
    if s == '1':
        return True
    return ZConfig.datatypes.stock_datatypes['boolean'](s)

datatypes['boolean'] = asBoolean

class Resolver(object):
    _bool_args = ()
    _int_args = ()
    _string_args = ()
    _bytesize_args = ()
    _float_args = ()
    _tuple_args = ()

    def interpret_kwargs(self, kw):
        unused = kw.copy()
        new = {}

        converters = [
            (datatypes['boolean'], self._bool_args),
            (datatypes['integer'], self._int_args),
            (datatypes['string'], self._string_args),
            (datatypes['byte-size'], self._bytesize_args),
            (datatypes['float'], self._float_args),
            (datatypes['string-list'], self._tuple_args),
        ]
        for convert, arg_names in converters:
            for arg_name in arg_names:
                value = unused.pop(arg_name, None)
                if value is not None:
                    value = convert(value)
                    new[arg_name] = value

        return new, unused


# Not a real resolver, but we use interpret_kwargs
class PostgreSQLAdapterHelper(Resolver):
    _int_args = ('connect_timeout', )
    _string_args = ('ssl_mode', )

    def __call__(self, parsed_uri, kw):
        dsn_args = [
            ('dbname', parsed_uri.path[1:]),
            ('user', parsed_uri.username),
            ('password', parsed_uri.password),
            ('host', parsed_uri.hostname),
            ('port', parsed_uri.port)
        ]

        kw, unused = self.interpret_kwargs(kw)
        dsn_args.extend(kw.items())

        dsn = ' '.join("%s='%s'" % arg for arg in dsn_args)

        def factory(options):
            from relstorage.adapters.postgresql import PostgreSQLAdapter
            return PostgreSQLAdapter(dsn=dsn, options=options)
        return factory, unused


class MySQLAdapterHelper(Resolver):
    _int_args = ('connect_timeout', 'client_flag', 'load_infile', 'compress',
                 'named_pipe')
    _string_args = ('unix_socket', 'init_command', 'read_default_file',
                    'read_default_group')

    def __call__(self, parsed_uri, kw):
        settings = {
            'db': parsed_uri.path[1:],
            'user': parsed_uri.username,
            'passwd': parsed_uri.password,
            'host': parsed_uri.hostname,
            'port': parsed_uri.port
        }

        kw, unused = self.interpret_kwargs(kw)
        settings.update(kw)

        def factory(options):
            from relstorage.adapters.mysql import MySQLAdapter
            return MySQLAdapter(options=options, **settings)
        return factory, unused


class OracleAdapterHelper(Resolver):
    _int_args = ('twophase', )
    _string_args = ('user', 'password', 'dsn')

    def __call__(self, parsed_uri, kw):
        kw, unused = self.interpret_kwargs(kw)

        def factory(options):
            from relstorage.adapters.oracle import OracleAdapter
            return OracleAdapter(options=options, **kw)
        return factory, unused

class SqliteAdapterHelper(Resolver):
    _string_args = ('path',)

    def __call__(self, parsed_uri, kw):
        kw, unused = self.interpret_kwargs(kw)

        def factory(options):
            from relstorage.adapters.sqlite.adapter import Sqlite3Adapter
            return Sqlite3Adapter(options=options, **kw)
        return factory, unused

# The relstorage support is inspired by django-zodb.

class RelStorageURIResolver(Resolver):
    _bool_args = (
        'read_only',
        'keep_history',
        'pack_gc',
        'pack_dry_run',
        'shared_blob_dir',
        'demostorage',
    )
    _int_args = (
        'cache_local_mb',
        'commit_lock_timeout',
        'commit_lock_id',
    )
    _string_args = (
        'name', 'blob_dir', 'replica_conf',
        'cache_module_name', 'cache_prefix',
        'cache_delta_size_limit', 'cache_local_compression',
        'driver',
    )
    _bytesize_args = (
        'blob_cache_size', 'blob_cache_size_check',
        'blob_cache_chunk_size',
        'cache_local_object_max',
    )
    _float_args = ('replica_timeout', 'pack_batch_timeout',
                   'pack_duty_cycle', 'pack_max_delay')
    _tuple_args = ('cache_servers',)

    def __init__(self, adapter_helper):
        self.adapter_helper = adapter_helper

    def __call__(self, uri):
        uri = uri.replace('postgres://', 'http://', 1)
        uri = uri.replace('mysql://', 'http://', 1)
        uri = uri.replace('oracle://', 'http://', 1)
        uri = uri.replace('sqlite://', 'http://', 1)
        parsed_uri = urlparse.urlsplit(uri)
        kw = dict(parse_qsl(parsed_uri.query))

        adapter_factory, kw = self.adapter_helper(parsed_uri, kw)
        kw, unused = self.interpret_kwargs(kw)

        demostorage = kw.pop('demostorage', False)
        options = Options(**kw)

        def factory():
            adapter = adapter_factory(options)
            storage = RelStorage(adapter=adapter, options=options)
            return storage if not demostorage else DemoStorage(base=storage)
        return factory, unused

postgresql_resolver = RelStorageURIResolver(PostgreSQLAdapterHelper())
mysql_resolver = RelStorageURIResolver(MySQLAdapterHelper())
oracle_resolver = RelStorageURIResolver(OracleAdapterHelper())
sqlite_resolver = RelStorageURIResolver(SqliteAdapterHelper())



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.adapters.sql.tests import test_sql

from .. import dialect

class TestSQLiteUpsertDialect(test_sql.TestUpsert):
    keep_history = False
    dialect = dialect.Sqlite3Dialect()
    dialect.compiler_class = lambda: dialect._Sqlite3UpsertCompiler

    def test_default(self):
        self.assertEqual(
            dialect.SQ3_SUPPORTS_UPSERT,
            dialect.SqliteCompiler is dialect._Sqlite3UpsertCompiler)

    insert_or_replace = test_sql.TestUpsert.insert_or_replace.replace(
        '%s', '?'
    )

    insert_or_replace_subquery = test_sql.TestUpsert.insert_or_replace_subquery.replace(
        '%s', '?'
    )

    upsert_unconstrained_subquery = test_sql.TestUpsert.upsert_unconstrained_subquery.replace(
        '%s', '?'
    ).replace(
        ' ON ', ' WHERE true ON '
    )


class TestSQLiteOldDialect(test_sql.TestUpsert):
    keep_history = False
    dialect = dialect.Sqlite3Dialect()
    dialect.compiler_class = lambda: dialect._Sqlite3NoUpsertCompiler

    insert_or_replace = (
        'INSERT OR REPLACE INTO object_state(zoid, state, tid, state_size) VALUES (?, ?, ?, ?)'
    )

    insert_or_replace_subquery = (
        'INSERT OR REPLACE INTO object_state(zoid, tid, state, state_size) '
        'SELECT zoid, ?, state, COALESCE(LENGTH(state), 0) FROM temp_store ORDER BY zoid'
    )

    upsert_unconstrained_subquery = (
        'INSERT OR REPLACE INTO object_state(zoid, tid, state, state_size) '
        'SELECT zoid, ?, state, COALESCE(LENGTH(state), 0) FROM temp_store'
    )



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides
from zope.schema.interfaces import IObject

from relstorage.options import Options
from relstorage.tests import TestCase

from .. import interfaces

class AdapterTestBase(TestCase):

    driver = 'auto'

    def _makeOne(self, options):
        raise NotImplementedError

    def test_implements(self):
        options = Options(driver=self.driver)
        adapter = self._makeOne(options)
        assert_that(adapter, validly_provides(interfaces.IRelStorageAdapter))

        # pylint:disable=no-value-for-parameter
        for attr_name, field in interfaces.IRelStorageAdapter.namesAndDescriptions():
            if IObject.providedBy(field):
                attr_iface = field.schema
                attr_val = getattr(adapter, attr_name)
                # Look for functions/methods in its dict and check for __wrapped__ attributes;
                # these come from perfmetrics @metricmethod and that breaks validation.
                for k in dir(type(attr_val)):
                    if k not in attr_iface:
                        continue
                    v = getattr(attr_val, k)
                    if callable(v) and hasattr(v, '__wrapped__'):
                        orig = v.__wrapped__
                        if hasattr(orig, '__get__'):
                            # Must be Python 3, we got the raw unbound function, we need to bind it
                            # and add the self parameter.
                            orig = orig.__get__(type(attr_val), attr_val)
                        try:
                            setattr(attr_val, k, orig)
                        except AttributeError:
                            # Must be slotted.
                            continue
                assert_that(attr_val, validly_provides(attr_iface))


def test_suite():
    # Nothing to see here.
    import unittest
    return unittest.TestSuite()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Interfaces provided by RelStorage database adapters"""
from __future__ import absolute_import

from ZODB.POSException import StorageError
from ZODB.POSException import ReadConflictError
from ZODB.POSException import ConflictError
from zope.interface import Attribute
from zope.interface import Interface

# pylint:disable=inherit-non-class,no-method-argument,no-self-argument
# pylint:disable=too-many-ancestors,too-many-lines

from relstorage.interfaces import Tuple
from relstorage.interfaces import Object
from relstorage.interfaces import Bool
from relstorage.interfaces import Factory
from relstorage.interfaces import IException

###
# Abstractions to support multiple databases.
###

class IDBDialect(Interface):
    """
    Handles converting from our internal "standard" SQL queries to
    something database specific.
    """

    # TODO: Fill this in.

class IDBDriver(Interface):
    """
    An abstraction over the information needed for RelStorage to work
    with an arbitrary DB-API driver.
    """

    __name__ = Attribute("The name of this driver")

    disconnected_exceptions = Tuple(
        description=(u"A tuple of exceptions this driver can raise on any operation if it is "
                     u"disconnected from the database."),
        value_type=Factory(IException)
    )

    close_exceptions = Tuple(
        description=(u"A tuple of exceptions that we can ignore when we try to "
                     u"close the connection to the database. Often this is the same "
                     u"or an extension of `disconnected_exceptions`."
                     u"These exceptions may also be ignored on rolling back the connection, "
                     u"if we are otherwise completely done with it and prepared to drop it. "),
        value_type=Factory(IException),
    )

    lock_exceptions = Tuple(
        description=u"A tuple of exceptions",
        value_type=Factory(IException),
    ) # XXX: Document

    use_replica_exceptions = Tuple(
        description=(u"A tuple of exceptions raised by connecting "
                     u"that should cause us to try a replica."),
        value_type=Factory(IException)
    )

    Binary = Attribute("A callable.")

    dialect = Object(IDBDialect, description=u"The IDBDialect for this driver.")

    cursor_arraysize = Attribute(
        "The value to assign to each new cursor's ``arraysize`` attribute.")

    supports_64bit_unsigned_id = Attribute("Can the driver handle 2**64 as a parameter?")

    connect = Attribute("""
    A callable to create and return a new connection object.

    The signature is not specified here because the
    required parameters differ between databases and drivers. The interface
    should be agreed upon between the :class:`IConnectionManager` and
    the drivers for its database.

    This connection, and all objects created from it such as cursors,
    should be used within a single thread only.
    """)

    def cursor(connection, server_side=False):
        """
        Create and return a new cursor sharing the state of the given
        *connection*.

        The cursor should be closed when it is no longer needed. The
        cursor should be considered forward-only (no backward
        scrolling) and ephemeral (results go away when the attached
        transaction is committed or rolled back).

        For compatibility, previous cursors should not have
        outstanding results pending when this is called and while the
        returned cursor is used (not all drivers permit multiple
        active cursors).

        If *server_side* is true (not the default), request that the
        driver creates a cursor that will **not** buffer the complete
        results of a query on the client. Instead, the results should
        be streamed from the server in batches. This can reduce the
        maximum amount of memory needed to handle results, if done
        carefully.

        For compatibility, server_side cursors can only be used
        to execute a single query.

        Most drivers (``psycopg2``, ``psycopg2cffi``, ``pg8000``,
        ``mysqlclient``) default to buffering the entire results
        client side before returning from the ``execute`` method. This
        can reduce latency and increase overall throughput, but at the
        cost of memory, especially if the results will be copied into
        different data structures.

        Not all drivers support server-side cursors; they will ignore
        that request. At this writing, this includes ``pg8000``. Some
        drivers (at this writing, only ``gevent MySQLdb``) always use
        server-side cursors. The ``cx_Oracle`` driver is unevaluated.

        ``psycopg2`` and ``psycopg2cffi`` both iterate in chunks of
        ``cur.itersize`` by default. PyMySQL seems to iterate one row at a time.
        ``mysqlclient`` defaults to also iterating one row at a time, but
        we patch that to operate in chunks of ``cur.arraysize``.
        """

    def binary_column_as_state_type(db_column_data):
        """
        Turn *db_column_data* into something that's a valid pickle
        state.

        Valid pickle states should be acceptable to
        `io.BytesIO` and `pickle.UnPickler`.

        *db_column_dat* came from a column of data declared to be of the
        type that we store state information in (e.g., a BLOB on MySQL
        or Oracle).
        """

    def binary_column_as_bytes(db_column_data):
        """
        Turn *db_column_data* into a `bytes` object.

        Use this when the specific type must be known,
        for example to prefix or suffix additional byte values
        like that produced by `p64`.
        """

    def enter_critical_phase_until_transaction_end(connection, cursor):
        """
        Given a connection and cursor opened by this driver, cause it
        to attempt to raise its priority and return results faster.

        This mostly has meaning for gevent drivers, which may limit
        the amount of time they spend in the hub and the number of context
        switches to other greenlets.

        This phase continues until *after* the ultimate call that
        commits or aborts is sent, but should revert to normal as quickly as
        possible after that. :class:`IRelStorageAdapter` may cooperate with
        the driver using implementation-specific methods to end the phase
        at an appropriate time if there is a hidden commit.

        This method must be idempotent (have the same effect if called more than
        once) within a given transaction.
        """

    def is_in_critical_phase(connection, cursor):
        """
        Answer whether :meth:`enter_critical_phase_until_transaction_end` is in effect.
        """

    def exit_critical_phase(connection, cursor):
        "If currently in a critical phase, de-escalate."

class IDBDriverSupportsCritical(IDBDriver):
    """
    A marker for database drivers that support
    critical phases.

    They promise that :meth:`enter_critical_phase_until_transaction_end`
    will do something useful.
    """

class IDBDriverFactory(Interface):
    """
    Information about, and a way to get, an `IDBDriver`
    implementation.
    """

    driver_name = Attribute("The name of this driver produced by this factory.")

    def check_availability():
        """
        Return a boolean indicating whether a call to this factory
        will return a driver (True) or will raise an error (False).
        """

    def __call__(): # pylint:disable=signature-differs
        """
        Return a new `IDBDriver` as represented by this factory.

        If it is not possible to do this, for example because the
        module cannot be imported, raise an `DriverNotAvailableError`.
        """

class DriverNotAvailableError(Exception):
    """
    Raised when a requested driver isn't available.
    """

    #: The name of the requested driver
    driver_name = None

    #: The `IDBDriverOptions` that was asked for the driver.
    driver_options = None

    #: The underlying reason string, for example, from an import error
    #: if such is available.
    reason = None

    def __init__(self, driver_name, driver_options=None, reason=None):
        super(DriverNotAvailableError, self).__init__(driver_name)
        self.driver_name = driver_name
        self.driver_options = driver_options
        self.reason = reason

    def _format_drivers(self):
        driver_factories = getattr(self.driver_options,
                                   'known_driver_factories',
                                   lambda: ())()
        driver_factories = list(driver_factories)
        if not driver_factories:
            return ''

        formatted = ' '.join(
            '%r (Module: %r; Available: %s)' % (
                factory.driver_name,
                # This attribute isn't in the interface,
                # it's an extension from AbstractModuleDriver
                getattr(factory, 'MODULE_NAME', '<unknown>'),
                factory.check_availability()
            )
            for factory in driver_factories
        )
        return ' Options: %s.' % (formatted,)

    def __str__(self):
        if self.reason:
            reason = ' (reason=%s)' % (self.reason,)
        else:
            reason = ''

        return '%s: Driver %r is not available%s.%s' % (
            type(self).__name__, self.driver_name,
            reason,
            self._format_drivers()
        )

    __repr__ = __str__


class UnknownDriverError(DriverNotAvailableError):
    """
    Raised when a driver that isn't registered at all is requested.
    """


class NoDriversAvailableError(DriverNotAvailableError):
    """
    Raised when there are no drivers available.
    """

    def __init__(self, driver_name='auto', driver_options=None, reason=None):
        super(NoDriversAvailableError, self).__init__(driver_name, driver_options, reason)


class IDBDriverOptions(Interface):
    """
    Implemented by a module to provide alternative drivers.
    """

    database_type = Attribute("A string naming the type of database. Informational only.")

    def select_driver(driver_name=None):
        """
        Choose and return an `IDBDriver`.

        The *driver_name* of "auto" is equivalent to a *driver_name* of
        `None` and means to choose the highest priority available driver.
        """

    def known_driver_factories():
        """
        Return an iterable of the potential `IDBDriverFactory`
        objects that can be used by `select_driver`.

        Each driver factory may or may not be available.

        The driver factories are returned in priority order, with the highest priority
        driver being first.
        """


###
# Creating and managing DB-API 2.0 connections.
# (https://www.python.org/dev/peps/pep-0249/)
###

class IConnectionManager(Interface):
    """
    Open and close database connections.

    This is a low-level interface; most operations should instead
    use a pre-existing :class:`IManagedDBConnection`.
    """

    isolation_load = Attribute("Default load isolation level.")
    isolation_store = Attribute("Default store isolation level.")
    isolation_read_committed = Attribute("Read committed.")
    isolation_serializable = Attribute("Serializable.")


    def open(
            isolation=None,
            deferrable=False,
            read_only=False,
            replica_selector=None,
            application_name=None,
            **kwargs):
        """Open a database connection and return (conn, cursor)."""

    def close(conn=None, cursor=None):
        """
        Close a connection and cursor, ignoring certain errors.

        Return a True value if the connection was closed cleanly;
        return a false value if an error was ignored.
        """

    def rollback_and_close(conn, cursor):
        """
        Rollback the connection and close it, ignoring certain errors.

        Certain database drivers, such as MySQLdb using the SSCursor, require
        all cursors to be closed before rolling back (otherwise it generates a
        ProgrammingError: 2014 "Commands out of sync").
        This method abstracts that.

        :return: A true value if the connection was closed without ignoring any exceptions;
            if an exception was ignored, returns a false value.
        """

    def rollback(conn, cursor):
        """
        Like `rollback_and_close`, but without the close, and letting
        errors pass.

        If an error does happen, then the connection and cursor are closed
        before this method returns.
        """

    def rollback_quietly(conn, cursor):
        """
        Like `rollback_and_close`, but without the close.

        :return: A true value if the connection was rolled back without ignoring any exceptions;
            if an exception was ignored, returns a false value (and the connection and cursor
            are closed before this method returns).
        """

    def begin(conn, cursor):
        """
        Call this on a store connection after restarting it.

        This lets the store connection know that it may need to begin a
        transaction, even if it was freshly opened.
        """

    def open_and_call(callback):
        """Call a function with an open connection and cursor.

        If the function returns, commits the transaction and returns the
        result returned by the function.
        If the function raises an exception, aborts the transaction
        then propagates the exception.
        """

    def open_for_load():
        """
        Open a connection for loading objects.

        This connection is read only, and presents a consistent view
        of the database as of the time the first statement is
        executed. It should be opened in ``REPEATABLE READ`` or higher
        isolation level. It must not be in autocommit.

        :return: ``(conn, cursor)``
        """

    def restart_load(conn, cursor, needs_rollback=True):
        """
        Reinitialize a connection for loading objects.

        This gets called when polling the database, so it needs to be
        quick.

        Raise one of self.disconnected_exceptions if the database has
        disconnected.
        """

    def open_for_store(**open_args):
        """
        Open and initialize a connection for storing objects.

        This connection is read/write, and its view of the database
        needs to be consistent for each statement, but should read a
        fresh snapshot on each statement for purposes of conflict
        resolution and cooperation with other store connections. It
        should be opened in ``READ COMMITTED`` isolation level,
        without autocommit. (Opening in ``REPEATABLE READ`` or higher,
        with a single snapshot, could reduce the use of locks, but
        increases the risk of serialization errors and having
        transactions rollback; we could handle that by raising
        ``ConflictError`` and letting the application retry, but only
        if we did that before ``tpc_finish``, and not all test cases
        can handle that either.)

        This connection will take locks on rows in the state tables,
        and hold them during the commit process.

        A connection opened by this method is the only type of
        connection that can hold the commit lock.

        :return: ``(conn, cursor)``
        """

    def restart_store(conn, cursor, needs_rollback=True):
        """
        Rollback and reuse a store connection.

        Raise one of self.disconnected_exceptions if the database
        has disconnected.

        You can set *needs_rollback* to false if you're certain
        the connection does not need rolled back.
        """

    def open_for_pre_pack():
        """
        Open a connection to be used for the pre-pack phase.

        This connection will make many different queries; each one
        must be consistent unto itself, but they do not all have to be
        consistent with each other. This is because the *first* query
        this object makes establishes a base state, and we will
        manually discard later changes seen in future queries.

        It will read from the state tables and write to the pack tables;
        it will not write to the state tables, nor hold the commit lock.
        It may hold share locks on state rows temporarily.

        This connection may be open for a long period of time, and
        will be committed as appropriate between queries. It is
        acceptable for this connection to be in autocommit mode, if
        required, but it is preferred for it not to be. This should be
        opened in ``READ COMMITTED`` isolation level.

        :return: ``(conn, cursor)``
        """

    def open_for_pack_lock():
        """
        Open a connection to be used for the sole purpose of holding
        the pack lock.

        Use a private connection (lock_conn and lock_cursor) to hold
        the pack lock. Have the adapter open temporary connections
        to do the actual work, allowing the adapter to use special
        transaction modes for packing, and to commit at will without
        losing the lock.

        If the database doesn't actually use a pack lock,
        this may return ``(None, None)``.
        """

    def cursor_for_connection(conn):
        """
        If the cursor returned by an open method was discarded
        for state management purposes, use this to get a new cursor.
        """

    def add_on_store_opened(f):
        """
        Add a callable(cursor, restart=bool) for when a store connection
        is opened.

        .. versionadded:: 2.1a1
        """

    def add_on_load_opened(f):
        """
        Add a callable (cursor, restart=bool) for when a load connection is opened.

        .. versionadded:: 2.1a1
        """


class IManagedDBConnection(Interface):
    """
    A managed DB connection consists of a DB-API ``connection`` object
    and a single DB-API ``cursor`` from that connection.

    This encapsulates proper use of ``IConnectionManager``, including
    handling disconnections and re-connecting at appropriate times.

    It is not allowed to use multiple cursors from a connection at the
    same time; not all drivers properly support that.

    If the DB-API connection is not open, presumed to be good, and
    previously accessed, this object has a false value.

    "Restarting" a connection means to bring it to a current view of
    the database. Typically this means a rollback so that a new
    transaction can begin with a new MVCC snapshot.
    """

    cursor = Attribute("The DB-API cursor to use. Read-only.")
    connection = Attribute("The DB-API connection to use. Read-only.")

    def __bool__():
        """
        Return true if the database connection is believed to be ready to use.
        """

    def __nonzero__():
        """
        Same as __bool__ for Python 2.
        """

    def drop():
        """
        Unconditionally drop (close) the database connection.
        """

    def rollback_quietly():
        """
        Rollback the connection and return a true value on success.

        When this completes, the connection will be in a neutral state,
        not idle in a transaction.

        If an error occurs during rollback, the connection is dropped
        and a false value is returned.
        """

    def isolated_connection():
        """
        Context manager that opens a new, distinct connection and
        returns its cursor.

        No matter what happens in the ``with`` block, the connection will be
        dropped afterwards.
        """

    def restart_and_call(f, *args, **kw):
        """
        Restart the connection (roll it back) and call a function
        after doing this.

        This may drop and re-connect the connection if necessary.

        :param callable f:
            The function to call: ``f(conn, cursor, *args, **kwargs)``.
            May be called up to twice if it raises a disconnected exception
            on the first try.

        :return: The return value of ``f``.
        """

    def enter_critical_phase_until_transaction_end():
        """
        As for :meth:`IDBDriver.enter_critical_phase_until_transaction_end`.
        """

class IManagedLoadConnection(IManagedDBConnection):
    """
    A managed connection intended for loading.
    """

class IManagedStoreConnection(IManagedDBConnection):
    """
    A managed connection intended for storing data.
    """

class IReplicaSelector(Interface):
    """Selects a database replica"""

    def current():
        """Get the current replica.

        Return a string.  For PostgreSQL and MySQL, the string is
        either a host:port specification or host name.  For Oracle,
        the string is a DSN.
        """

    def next():
        """Return the next replica to try.

        Return None if there are no more replicas defined.
        """




class IDatabaseIterator(Interface):
    """Iterate over the available data in the database"""

    def iter_objects(cursor, tid):
        """Iterate over object states in a transaction.

        Yields (oid, prev_tid, state) for each object state.
        """

    def iter_transactions(cursor):
        """
        Iterate over the transaction log, newest first.

        Skips packed transactions. Yields (tid, username, description,
        extension) for each transaction.
        """

    def iter_transactions_range(cursor, start=None, stop=None):
        """
        Return an indexable object over the transactions in the given range, oldest
        first.

        Includes packed transactions.

        Has an object with the properties ``tid_int``, ``username``
        (bytes) ``description`` (bytes) ``extension`` (bytes) and
        ``packed`` (boolean) for each transaction.
        """

    def iter_object_history(cursor, oid):
        """
        Iterate over an object's history.

        Yields an object with the properties ``tid_int``, ``username``
        (bytes) ``description`` (bytes) ``extension`` (bytes) and
        ``pickle_size`` (int) for each transaction.

        :raises KeyError: if the object does not exist
        """

    def iter_current_records(cursor, start_oid_int=0):
        """
        Cause the *cursor* (which should be a server-side cursor)
        to execute a query that will iterate over
        ``(oid_int, tid_int, state_bytes)`` values for all the current objects.

        Each current object is returned only once, at the transaction most recently
        committed for it.

        Returns a generator.

        For compatibility with FileStorage, this must iterate in ascending
        OID order; it must also accept an OID to begin with for compatibility with
        zodbupdate.
        """

class ILocker(Interface):
    """Acquire and release the commit and pack locks."""

    def lock_current_objects(cursor, read_current_oid_ints, shared_locks_block):
        """
        Lock the objects being modified in the current transaction
        exclusively, plus the relevant rows for the objects whose OIDs
        are contained in *read_current_oid_ints* with a read lock.

        The exclusive locks should always be taken in a blocking fashion;
        the shared read locks should be taken without blocking (raising an
        exception if blocking would occur) if possible, unless *shared_locks_block*
        is set to True.

        See :meth:`IRelStorageAdapter.lock_objects_and_detect_conflicts`
        for a description of the expected behaviour.

        This should be done as part of the voting phase of TPC, before
        taking out the final commit lock.

        Returns nothing.

        Typically this will be followed by a call to
        :meth:`detect_conflict`.
        """

    def hold_commit_lock(cursor, ensure_current=True, nowait=False):
        """
        Acquire the commit lock.

        If *ensure_current* is True (the default), other tables may be
        locked as well, to ensure the most current data is available.
        When using row level locks, *ensure_current* is always
        implicit.

        With *nowait* set to True, only try to obtain the lock without
        waiting and return a boolean indicating if the lock was
        successful. **Note:** this parameter is deprecated and will be removed
        in the future; it is not currently used.

        Should raise `UnableToAcquireCommitLockError` if the lock can not
        be acquired before a configured timeout.
        """

    def release_commit_lock(cursor):
        """Release the commit lock"""

    def hold_pack_lock(cursor):
        """Try to acquire the pack lock.

        Raise UnableToAcquirePackUndoLockError if packing or undo is already in progress.
        """

    def release_pack_lock(cursor):
        """Release the pack lock."""


class IObjectMover(Interface):
    """Move object states to/from the database and within the database."""

    def load_current(cursor, oid):
        """
        Returns the current state and integer tid for an object.

        *oid* is an integer. Returns (None, None) if object does not
        exist.
        """

    def load_currents(cursor, oids):
        """
        Returns the oid integer, state, and integer tid for all the specified
        objects.

        *oids* is an iterable of integers. If any objects do no exist,
        they are ignored.
        """

    def load_revision(cursor, oid, tid):
        """Returns the state for an object on a particular transaction.

        Returns None if no such state exists.
        """

    def exists(cursor, oid):
        """Returns a true value if the given object exists."""

    def load_before(cursor, oid, tid):
        """Returns the state and tid of an object before transaction tid.

        Returns (None, None) if no earlier state exists.
        """

    def get_object_tid_after(cursor, oid, tid):
        """Returns the tid of the next change after an object revision.

        Returns None if no later state exists.
        """

    def current_object_tids(cursor, oids, timeout=None):
        """
        Returns the current ``{oid_int: tid_int}`` for specified object ids.

        Note that this may be a BTree mapping, not a dictionary.

        :param oids: An iterable of OID integers.
        :keyword float timeout: If not None, this is an approximate upper bound
           (in seconds) on how long this function will run.
        :raises AggregateOperationTimeoutError: If the timeout was exceeded.
           This will have one extra attribute set, ``partial_result``, which will be a
           (partial) mapping of the results collected before the timeout.
        """

    def on_store_opened(cursor, restart=False):
        """Create the temporary table for storing objects.

        This method may be None, meaning no store connection
        initialization is required.
        """

    def make_batcher(cursor, row_limit):
        """Return an object to be used for batch store operations.

        *row_limit* is the maximum number of rows to queue before
        calling the database.
        """

    def store_temps(cursor, state_oid_tid_iter):
        """
        Store many objects in the temporary table.

        *batcher* is an object returned by :meth:`make_batcher`.

        *state_oid_tid_iter* is an iterable providing tuples
        ``(data, oid_int, prev_tid_int)``. It is guaranteed that the
        ``oid_int`` values will be distinct. It is further guaranteed that
        this method will not be called more than once in a given transaction;
        further updates to the temporary table will be made using
        ``replace_temps``, which is also only called once.
        """

    def restore(cursor, batcher, oid, tid, data):
        """Store an object directly, without conflict detection.

        Used for copying transactions into this database.

        batcher is an object returned by self.make_batcher().
        """

    def detect_conflict(cursor):
        """
        Find all conflicts in the data about to be committed (as stored
        by :meth:`store_temps`)

        Returns a sequence of
        ``(oid, committed_tid, attempted_committed_tid, committed_state)`` where
        each entry refers to a conflicting object. The *committed_state* **must** be
        returned.

        This method should be called during the ``tpc_vote`` phase of a transaction,
        with :meth:`ILocker.lock_current_objects` held.
        """

    def replace_temps(cursor, state_oid_tid_iter):
        """
        Replace all objects in the temporary table with new data from
        *state_oid_tid_iter*.

        This happens after conflict resolution. The param is as for
        ``store_temps``.

        Implementations should try to perform this in as few database operations
        as possible.
        """

    def move_from_temp(cursor, tid, txn_has_blobs):
        """
        Move the temporarily stored objects to permanent storage.

        *tid* is the integer tid of the transaction being committed.

        Returns nothing.

        The steps should be as follows:

            - If we are preserving history, then ``INSERT`` into
              ``object_state`` the values stored in ``temp_store``,
              remembering to coalesce the
              ``LENGTH(temp_store.state)``.

            - Otherwise, when we are not preserving history,
              ``INSERT`` missing rows from ``object_state`` into
              ``temp_store``, and ``UPDATE`` rows that were already
              there. (This is best done with an upsert). If blobs are
              involved, then ``DELETE`` from ``blob_chunk`` where the
              OID is in ``temp_store``.

            - For both types of storage, ``INSERT`` into
              ``blob_chunk`` the values from ``temp_blob_chunk``. In a
              history-free storage, this may be combined with the last
              step in an ``UPSERT``.
        """

    def update_current(cursor, tid):
        """
        Update the current object pointers.

        *tid* is the integer tid of the transaction being committed.

        Returns nothing. This does nothing when the storage is history
        free.

        When the storage preserves history, all the objects in
        ``object_state`` having the given *tid* should have their
        (oid, *tid*) stored into ``current_object``. This can be done
        with a single upsert.

        XXX: Why do we need to look at ``object_state``? Is there a
        reason we can't look at the smaller ``temp_store``? Conflict
        resolution maybe?
        """

    def download_blob(cursor, oid, tid, filename):
        """Download a blob into a file.

        Returns the size of the blob file in bytes.
        """

    def upload_blob(cursor, oid, tid, filename):
        """Upload a blob from a file.

        If tid is None, upload to the temporary table.
        """


class IOIDAllocator(Interface):
    """
    Allocate OIDs and control future allocation.

    The cursor passed here must be from a
    :meth:`store connection <IConnectionManager.open_for_store>`.
    """

    def new_oids(cursor):
        """
        Return a new :class:`list` of new, unused integer OIDs.

        The list should be contiguous and must be in sorted order from
        highest to lowest. It must never contain 0.
        """

    def set_min_oid(cursor, oid_int):
        """
        Ensure the next OID (the rightmost value from
        :meth:`new_oids`) is greater than the given *oid_int*.
        """

    def reset_oid(cursor):
        """
        Cause the sequence of OIDs to begin again from the beginning.
        """


class IPackUndo(Interface):
    """Perform pack and undo operations"""

    MAX_TID = Attribute("The maximum TID that can be stored.")

    def verify_undoable(cursor, undo_tid):
        """Raise UndoError if it is not safe to undo the specified txn.
        """

    def undo(cursor, undo_tid, self_tid):
        """Undo a transaction.

        Parameters: "undo_tid", the integer tid of the transaction to undo,
        and "self_tid", the integer tid of the current transaction.

        Returns the states copied forward by the undo operation as a
        list of (oid, old_tid).

        May raise UndoError.
        """

    def fill_object_refs(conn, cursor, get_references):
        """Update the object_refs table by analyzing new transactions.
        """

    def choose_pack_transaction(pack_point):
        """Return the transaction before or at the specified pack time.

        Returns None if there is nothing to pack.
        """

    def pre_pack(pack_tid, get_references):
        """Decide what to pack.

        pack_tid specifies the most recent transaction to pack.

        get_references is a function that accepts a stored object state
        and returns a set of OIDs that state refers to.
        """

    def pack(pack_tid, packed_func=None):
        """Pack.  Requires the information provided by pre_pack.

        packed_func, if provided, will be called for every object state
        packed, just after the object is removed. The function must
        accept two parameters, oid and tid (64 bit integers).
        """

    def deleteObject(cursor, oid_int, tid_int):
        """
        Delete the revision of *oid_int* in transaction *tid_int*.

            This method marks an object as deleted via a new object
            revision. Subsequent attempts to load current data for the
            object will fail with a POSKeyError, but loads for
            non-current data will suceed if there are previous
            non-delete records. The object will be removed from the
            storage when all not-delete records are removed.

            The serial argument must match the most recently committed
            serial for the object. This is a seat belt.

            --- Documentation for ``IExternalGC``

        In history-free databases there is no such thing as a delete record, so
        this should remove the single
        revision of *oid_int* (which *should* be checked to verify it
        is at *tid_int*), leading all access to *oid_int* in the
        future to throw ``POSKeyError``.

        In history preserving databases, this means to set the state for the object
        at the transaction to NULL, signifying that it's been deleted. A subsequent
        pack operation is required to actually remove these deleted items.
        """

class IPoller(Interface):
    """Poll for new data"""

    def get_current_tid(cursor):
        """
        Returns the highest transaction ID visible to the cursor.

        If there are no transactions, returns 0.
        """

    def poll_invalidations(conn, cursor, prev_polled_tid):
        """
        Polls for new transactions.

        *conn* and *cursor* must have been created previously by
        ``open_for_load()`` (a snapshot connection). *prev_polled_tid*
        is the tid returned at the last poll, or None if this is the
        first poll.

        If the database has disconnected, this method should raise one
        of the exceptions listed in the disconnected_exceptions
        attribute of the associated IConnectionManager.

        Returns ``(changes, new_polled_tid)``, where changes is either
        an iterable of (oid, tid) that have changed, or None to
        indicate that the changes are too complex to list; this must
        cause local storage caches to be invalidated.
        ``new_polled_tid`` is never None.

        Important: You must consume the changes iterable, and you must
        not make any other queries until you do.

        This method may raise :class:`StaleConnectionError` (a
        ``ReadConflictError``) if the database has reverted to an
        earlier transaction, which can happen in an asynchronously
        replicated database. This exception is one that is transient
        and most transaction middleware will catch it and retry the
        transaction.
        """

class ISchemaInstaller(Interface):
    """Install the schema in the database, clear it, or uninstall it"""

    def prepare():
        """
        Create the database schema if it does not already exist.

        Perform any migration steps needed, and call :meth:`verify`
        before returning.
        """

    def verify():
        """
        Ensure that the schema that's installed can be used by this
        RelStorage.

        If it cannot, for example it's history-preserving and we were configured
        to be history-free, raise an exception.
        """

    def zap_all():
        """Clear all data out of the database."""

    def drop_all():
        """Drop all tables and sequences."""


class IScriptRunner(Interface):
    """Run database-agnostic SQL scripts.

    Using an IScriptRunner is appropriate for batch operations and
    uncommon operations that can be slow, but is not appropriate
    for performance-critical code.
    """

    script_vars = Attribute(
        """A mapping providing replacements for parts of scripts.

        Used for making scripts compatible with databases using
        different parameter styles.
        """)

    def run_script_stmt(cursor, generic_stmt, generic_params=()):
        """Execute a statement from a script with the given parameters.

        generic_params should be either an empty tuple (no parameters) or
        a map.

        The input statement is generic and will be transformed
        into a database-specific statement.
        """

    def run_script(cursor, script, params=()):
        """Execute a series of statements in the database.

        params should be either an empty tuple (no parameters) or
        a map.

        The statements are transformed by run_script_stmt
        before execution.
        """

    def run_many(cursor, stmt, items):
        """Execute a statement repeatedly.  Items should be a list of tuples.

        stmt should use '%s' parameter format (not %(name)s).
        """

    # Note: the Oracle implementation also provides run_lob_stmt, which
    # is useful for reading LOBs from the database quickly.


class ITransactionControl(Interface):
    """Begin, commit, and abort transactions."""

    def get_tid(cursor):
        """Returns the most recent tid."""

    def add_transaction(cursor, tid, username, description, extension,
                        packed=False):
        """Add a transaction."""

    def delete_transaction(cursor, tid):
        """Remove a transaction."""

    def commit_phase1(store_connection, tid):
        """
        Begin a commit. Returns the transaction name.

        The transaction name must not be None.

        This method should guarantee that :meth:`commit_phase2` will
        succeed, meaning that if commit_phase2() would raise any
        error, the error should be raised in :meth:`commit_phase1`
        instead.

        :param store_connection: An :class:`IManagedStoreConnection`
        """

    def commit_phase2(store_connection, txn, load_connection):
        """
        Final transaction commit.

        *txn* is the name returned by :meth:`commit_phase1`.

        :param store_connection: An :class:`IManagedStoreConnection`
           This is what must be committed.
        :param load_connection: An :class:`IManagedLoadConnection`
           corresponding to the store connection. If helpful to the database,
           (for example, for resource reasons) implementations may rollback
           the connection immediately before committing the store connection.
        """

    def abort(store_connection, txn=None):
        """
        Abort the commit, ignoring certain exceptions.

        If *txn* is not None, phase 1 is also aborted.

        :param store_connection: An :class:`IManagedStoreConnection`

        :return: A true value if the connection was rolled back
                 without ignoring any exceptions; if an exception was
                 ignored, returns a false value (and the connection
                 and cursor are closed before this method returns).
        """

class IDBStats(Interface):
    """
    Collecting, viewing and updating database information.
    """

    def get_object_count():
        """Returns the approximate number of objects in the database"""

    def get_db_size():
        """Returns the approximate size of the database in bytes"""

    def large_database_change():
        """
        Call this when the database has changed substantially,
        and it would be a good time to perform any updates or
        optimizations.
        """


class IRelStorageAdapter(Interface):
    """
    A database adapter for RelStorage.

    Historically, this has just been a holding place for other components
    for the particular database. However, it is moving to holding algorithms
    involved in the storage; this facilitates moving chunks of functionality
    into database stored procedures as appropriate. The basic algorithms are
    implemented in :class:`.adapter.AbstractAdapter`.

    """

    driver = Object(IDBDriver)
    connmanager = Object(IConnectionManager)
    dbiter = Object(IDatabaseIterator)
    keep_history = Bool(description=u"True if this adapter supports undo")
    locker = Object(ILocker)
    mover = Object(IObjectMover)
    oidallocator = Object(IOIDAllocator)
    packundo = Object(IPackUndo)
    poller = Object(IPoller)
    runner = Object(IScriptRunner)
    schema = Object(ISchemaInstaller)
    stats = Object(IDBStats)
    txncontrol = Object(ITransactionControl)

    def new_instance():
        """
        Return an instance for use by another RelStorage instance.

        Adapters that are stateless can simply return self. Adapters
        that have mutable state must make a clone and return it.
        """

    def release():
        """
        Release the resources held uniquely by this instance.
        """

    def close():
        """
        Release the resources held by this instance and all child instances.
        """

    def __str__():
        """Return a short description of the adapter"""

    def lock_database_and_choose_next_tid(cursor,
                                          username,
                                          description,
                                          extension):
        """
        Lock the database with the commit lock and allocate the next
        tid.

        In a simple implementation, this will first obtain the commit
        lock with around :meth:`ILocker.hold_commit_lock`. Then it
        will query the current most recently committed TID with
        :meth:`ITransactionControl.get_tid`. It will choose the next
        TID based on that value and the current timestamp, and then it
        will write that value to the database with
        :meth:`ITransactionControl.add_transaction`.

        The *username*, *description* and *extension* paramaters are as for
        ``add_transaction.

        :return: The new TID integer.
        """

    def lock_database_and_move(
            store_connection, load_connection,
            transaction_has_blobs,
            ude,
            commit=True,
            committing_tid_int=None,
            after_selecting_tid=None
    ):
        """
        Lock the database, choose the next TID, and move temporary
        data into its final place.

        This is used in two modes. In the usual case, *commit* will be
        true, and this method is called to implement the final step of
        ``tpc_finish``. In that case, this method is responsible for
        committing the transaction (using the *store_connection*
        provided). When *commit* is false, this method is effectively
        part of ``tpc_vote`` and must **not** commit.

        The *blobhelper* is the :class:`IBlobHelper`. This method is
        responsible for moving blob data into place if the blob data
        is stored on the server and there are blobs in this
        transaction. (Implementations that use stored procedures will
        probably not need this argument; it is here to be able to
        provide ``txn_has_blobs`` to
        :meth:`IObjectMover.move_from_temp`.)

        *ude* is a tuple of ``(username, description, extension)``.

        If *committing_tid_int* is None, then this method must lock
        the database and choose the next TID as if by calling
        :meth:`lock_database_and_choose_next_tid` (passing in the
        expanded *ude*); if it is **not** None, the database has
        already been locked and the TID selected.

        *after_selecting_tid* is a function of one argument, the
        committing integer TID. If it is proveded, it must be called
        once the TID has been selected and temporary data moved into
        place.

        Implementations are encouraged to do all of this work in as
        few calls to the database as possible with a stored procedure
        (ideally one call). The default implementation will use
        :meth:`lock_database_and_choose_next_tid`,
        :meth:`IObjectMover.move_from_temp`,
        :meth:`IObjectMover.update_current` and
        :meth:`ITransactionControl.commit_phase1` and
        :meth:`ITransactionControl.commit_phase2`.

        When committing, implementations are encouraged to exit any
        :meth:`critical phase <IDBDriver.enter_critical_phase_until_transaction_end>`
        in the most timely manner possible after ensuring that the commit
        request has been sent, especially if only one
        communication with the database is required that may block for an arbitrary
        time to get the lock.

        :param load_connection: The load connection corresponding to the store
            connection. When *commit* is true, this **may** be rolled back
            immediately before actually committing the *store_connection*, if
            that assists the database (for example, with resource management).
            If *commit* is not true, the load connection must not be changed.
            Rolling back the load connection is optional.

            Be careful to use the *store_connection* for all operations requiring
            a current view of the database or any writes. The load connection
            has a historical view and is not writable.

        :return: A tuple ``(committing_tid_int, prepared_txn_id)``;
                 the *prepared_txn_id* is irrelevant if *commit* was
                 true.
        """

    def lock_objects_and_detect_conflicts(
            cursor,
            read_current_oids,
    ):
        """
        Without taking the commit lock, lock the objects this
        transaction wants to modify (for exclusive update) and the
        objects in *read_current_oids* (for shared read).

        Returns an iterable of ``(oid_int, committed_tid_int,
        tid_this_txn_saw_int, committed_state)`` for current objects
        that were locked, plus objects which had conflicts.

        Implementations are encouraged to do all this work in as few
        calls to the database as possible with a stored procedure. The
        default implementation will use
        :meth:`ILocker.lock_current_objects`,
        :meth:`IObjectMover.current_object_tids`, and
        :meth:`IObjectMover.detect_conflicts`.

        This method may raise the same lock exceptions and
        :meth:`ILocker.lock_current_objects`. In particular, it should
        take care to distinguish between a failure to acquire an
        update lock and a failure to acquire a read lock by raising
        the appropriate exceptions
        (:class:`UnableToLockRowsToModifyError` and
        :class:`UnableToLockRowsToReadCurrentError`, respectively).

        Because two separate classes of locks are to be obtained,
        implementations will typically need to make two separate
        locking queries. If the second of those queries fails, the
        implementation is encouraged to immediately release locks
        taken by the first operation before raising an exception.
        Ideally this happens in the database stored procedure.
        (PostgreSQL works this way automatically --- any error rolls
        the transaction back and releases locks --- but this takes
        work on MySQL except in the case of deadlock.)

        .. rubric:: Read Current Locks

        The *read_current_oids* maps OID integers to expected TID
        integers. Each such object must be read share locked for the
        duration of the transaction so we can ensure that the final
        commit occurs without those objects having been changed. From
        ReadVerifyingStorage's ``checkCurrentSerialInTransaction``:
        "If no [ReadConflictError] exception is raised, then the
        serial must remain current through the end of the
        transaction."

        Applications sometimes (often) perform readCurrent() on the
        *wrong* object (for example: the ``BTree`` object or the
        ``zope.container`` container object, when what is really
        required, what will actually be modified, is a ``BTree``
        bucket---very hard to predict), so very often these objects
        will not ever be modified. (Indeed, ``BTree.__setitem__`` and
        ``__delitem__`` perform readCurrent on the BTree itself; but
        once the first bucket has been established, the BTree will not
        usually be modified.) A share lock is enough to prevent any
        modifications without causing unnecessary blocking if the
        object would never be modified.

        In the results, if the ``tid_this_txn_saw_int`` is ``None``,
        that was an object we only read, and share locked. If the
        ``committed_tid_int`` does not match the TID we expected to
        get, then the caller will raise a ``ReadConflictError`` and
        abort the transaction. The implementation is encouraged to
        return all such rows *first* so that these inexpensive checks
        can be accomplished before the more expensive conflict
        resolution process.

        Optionally, if this method can detect a read current violation
        based on the data in *read_current_oids* at the database
        level, it may raise a :class:`ReadConflictError`. This method
        is also allowed and encouraged to *only* return read current
        violations; further, it need only return the first such
        violation (because an exception will immediately be raised.)

        Implementations are encouraged to use the database ``NOWAIT``
        feature (or equivalent) to take read current locks. If such an
        object is already locked exclusively, that means it is being
        modified and this transaction is racing the modification
        transaction. Taking the lock with ``NOWAIT`` and raising an
        error lets the write transaction proceed, while this one rolls
        back and retries. (Hopefully the write transaction finishes
        quickly, or several retries may be needed.)

        .. rubric:: Modified Objects and Conflicts

        All objects that this transaction intends to modify (which are
        in the ``temp_store`` table) must be exclusively write locked
        when this method returns.

        The remainder of the results (where ``tid_this_txn_saw_int``
        is not ``None``) give objects that we have detected a conflict
        (a modification that has committed earlier to an object this
        transaction also wants to modify).

        As an optimization for conflict resolution,
        ``committed_state`` may give the current committed state of
        the object (corresponding to ``committed_tid_int``), but is
        allowed to be ``None`` if there isn't an efficient way to
        query that in bulk from the database.

        .. rubric:: Deadlocks, and Shared vs Exclusive Locks

        It might seem that, because no method of a transaction
        (*except* for ``restore()``) writes directly to the
        ``object_state`` or ``current_object`` table *before*
        acquiring the commit lock, a share lock is enough even for
        objects we're definitely going to modify. That way leads to
        deadlock, however. Consider this order of operations:

            1. Tx a: LOCK OBJECT 1 FOR SHARE. (Tx a will modify this.)

            2. Tx b: LOCK OBJECT 1 FOR SHARE. (Tx b is just
               readCurrent.)

            3. Tx a: Obtain commit lock.

            4. Tx b: attempt to obtain commit lock; block.

            5. Tx a: UPDATE OBJECT 1; attempt to escalate shared lock
               to exclusive lock. --> DEADLOCK with the shared lock
               from step 2.

        Tx a needs to raise the lock of object 1, but Tx b's share
        lock is preventing it. Meanwhile, Tx b wants the commit lock,
        but Tx a is holding it.

        If Tx a took an exclusive lock, it would either block Tx b
        from getting a share lock, or be blocked by Tx b's share lock;
        either way, whichever one got to the commit lock would be able
        to complete.

        Further, it is trivial to show that when using two lock
        classes, two transactions that have overlapping sets of
        objects (e.g., a wants shared on ``(1, 3, 5, 7)`` and
        exclusive on ``(2, 4, 6, 8)`` and b wants shared on ``(2, 4,
        6, 8)`` and exclusive on ``(3, 5, 7)``), can easily deadlock
        *before* taking the commit lock, no matter how we interleave
        those operations. This is true if they both take their
        exclusive locks first and then attempt share locks on the
        remainder, both take shared locks on everything and attempt to
        upgrade to exclusive on that subset, or both take just the
        shared locks and then attempt to take the exclusive locks.
        This extends to more than two processes.

        **That's perfectly fine.**

        As long as the database either supports ``NOWAIT``
        (immediately error when you fail to get a requested lock) or
        rapid deadlock detection resulting in an error, we can catch
        that error and turn it into the ``ReadConflictError`` it
        actually is.

        PostgreSQL supports ``NOWAIT`` (and deadlock detection, after
        a small but configurable delay). MySQL's InnoDB supports rapid
        deadlock detection, and starting with MySQL 8, it supports
        ``NOWAIT``.

        .. rubric:: Lock Order

        The original strategy was to first take exclusive locks of
        things we will be modifying. Once that succeeds, then we
        attempt shared locks of readCurrent using ``NOWAIT``. If that
        fails because we can't get a lock, we know someone is in the
        process of modifying it and we have a conflict. If we get the
        locks, we still have to confirm the TIDs are the things we
        expect. (A possible optimization is to do those two steps at
        once, in the database. ``SELECT FOR SHARE WHERE oid = X and
        TID = x``. If we don't get the right number of rows,
        conflict.) This prioritizes writers over readers: readers fail
        at the expense of writers.

        However, it means that if we're going to fail a transaction
        because an object we'd like to read lock has been modified, we
        have to wait until we timeout, or acquire all of our exclusive
        locks. Depending on transaction ordering, this could mean
        unnecessarily long delays. Suppose Tx a wants to write to
        objects 1 and 2, and Tx b wants to write to 1 but only read 2.
        (Recall that a ZODB connection automatically upgrades
        readCurrent() into just modifications. 1 and 2 could be bank
        accounts; Tx a is doing a transfer, but Tx b is checking
        collateral (account 2) for a loan that was approved and
        transferring that into 1.)

            1. Tx a: LOCK 1, 2 EXCLUSIVE. (No share locks.)

            2. Tx b: LOCK 1 EXCLUSIVE. -> Block; queue for lock 1.

            3. Tx a: Resolve a conflict in 1, wait for the commit
               lock, and finally commit and release the locks on 1 and

            4. Tx c: LOCK 2 EXCLUSIVE.

            5. Tx b: LOCK 2 SHARE -> wait exception.

        Here, Tx b had to wait while Tx a finished its entire
        business, only to have Tx c swoop in and get the lock first,
        leading to Tx b raising an exception. Some databases guarantee
        that locks are handed off in FIFO fashion, but not all do.
        Even if the database granted Tb b the share lock first, it
        would still discover that the TID had changed and raise an
        exception. Meanwhile, Tx b has been holding an exclusive lock
        on 1 this entire time, preventing anyone else from modifying
        it.

        If we take the share locks first, the scenario looks like
        this:

            1. Tx a: LOCK 1, 2 EXCLUSIVE. (No share locks.)

            2. Tx b: LOCK 2 SHARE -> wait exception; begin retry.

        Tx b gets much quicker notification that it won't be able to
        progress and begins a retry.

        Both orders have the problem that if Tx a takes longer to
        commit than Tx b does to begin its retry, Tx b may take the
        same action against the same state of the database several
        times in a row before giving up. What's different is that in
        the exclusive first version, that only happens if Tx a and Tx
        b have no exclusive locks in common:

            1. Tx a: Lock 1 EXCLUSIVE.

            2. Tx b: Lock 2 EXCLUSIVE.

            3. Tx b: Lock 1 SHARE -> wait exception, begin retry.

        However, this can be mitigated by introducing small backoff
        delays in the transaction retry logic.

        Suppose a different transaction already modified object 2.
        With the original lock order (exclusive first), the scenario
        doesn't change. Tx b has to wait for Tx a to finish before it
        can determine that fact (Tx a has to resolve a conflict or
        fail). If Tx b got to go first, it would relatively quickly
        discover this fact, but at the cost of waiting for an
        exclusive lock for an arbitrary amount of time:

            1. Tx b: Lock 1 EXCLUSIVE. (Potentially blocks.)

            2. Tx a: Lock 1, 2 EXCLUSIVE. -> Block; queue for lock 1.

            3. Tx b: Lock 2 SHARE.

            4. Tx b: Determine 2 has been modified; raise
               ReadConflictError.

        Taking the share lock first solves this concern; Tx b is
        immediately able to determine that 2 has been modified and
        quickly raise an exception without holding any other locks.

        :param cursor: The store cursor.
        :param read_current_oids: A mapping from oid integer to tid
            integer that the transaction expects.
        """


###
# Exceptions
###

class ReplicaClosedException(Exception):
    """The connection to the replica has been closed"""

class UnableToAcquireLockError(Exception):
    "A lock cannot be acquired."

class UnableToAcquireCommitLockError(StorageError, UnableToAcquireLockError):
    """
    The commit lock cannot be acquired due to a timeout.

    This means some other transaction had the lock we needed. Retrying
    the transaction may succeed.

    However, for historical reasons, this exception is not a ``TransientError``.
    """

# TransientError -> ConflictError -> ReadConflictError

class UnableToLockRowsToModifyError(ConflictError, UnableToAcquireLockError):
    """
    We were unable to lock one or more rows that we intend to modify
    due to a timeout.

    This means another transaction already had the rows locked,
    and thus we are in conflict with that transaction. Retrying the
    transaction may succeed.

    This is a type of ``ConflictError``, which is a transient error.
    """

class UnableToLockRowsToReadCurrentError(ReadConflictError, UnableToAcquireLockError):
    """
    We were unable to lock one or more rows that belong to an object
    that ``Connection.readCurrent()`` was called on.

    This means another transaction already had the rows locked with
    intent to modify them, and thus we are in conflict with that
    transaction. Retrying the transaction may succeed.

    This is a type of ``ReadConflictError``, which is a transient error.
    """

class UnableToAcquirePackUndoLockError(StorageError, UnableToAcquireLockError):
    """A pack or undo operation is in progress."""


class StaleConnectionError(ReadConflictError):
    """
    Raised by `IPoller.poll_invalidations` when a stale connection is
    detected.
    """

    @classmethod
    def from_prev_and_new_tid(cls, prev_polled_tid, new_polled_tid):
        return cls(
            "The database connection is stale: new_polled_tid=%d, "
            "prev_polled_tid=%d." % (new_polled_tid, prev_polled_tid))


class AggregateOperationTimeoutError(Exception):
    """
    Raised when an aggregate operation in RelStorage detects
    that it has exceeded a specified timeout.
    """

    #: If the function knows a useful, partial result, it may set this
    #: attribute. Check it against the class value to see if it's been set.
    partial_result = object()



##############################################################################
#
# Copyright (c) 2008,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

# pylint:disable=too-many-public-methods,unused-argument
from __future__ import absolute_import
from __future__ import print_function

import os

from relstorage._compat import PY3

from . import test_blobhelper
from .test_blobhelper import write_file
from .test_blobhelper import read_file
from .test_blobhelper import test_oid
from .test_blobhelper import test_tid



class CacheBlobHelperTest(test_blobhelper.BlobHelperTest):
    # Tests that only apply to cache dirs

    shared_blob_dir = False
    move_method = 'finish'

    def test_ctor_with_private_blob_dir(self):
        blobhelper = self._make_default()
        self.assertIsNotNone(blobhelper.fshelper)
        self.assertIsNotNone(blobhelper.cache_checker)
        from ZODB.blob import LAYOUTS
        self.assertEqual(blobhelper.fshelper.layout, LAYOUTS['zeocache'])

    def test_new_instance(self):
        class DummyAdapter2(object):
            pass

        blobhelper = self._make_default()
        blobhelper2 = blobhelper.new_instance(DummyAdapter2())
        # The adapter we passed is assigned, not the one the original object
        # had cached.
        self.assertNotIsInstance(blobhelper.adapter, DummyAdapter2)
        self.assertIsInstance(blobhelper2.adapter, DummyAdapter2)
        self.assertIs(blobhelper.fshelper, blobhelper2.fshelper)
        self.assertIs(blobhelper.cache_checker, blobhelper2.cache_checker)


    def test_download_found(self):
        fn = os.path.join(self.blob_dir, '0001')
        blobhelper = self._make_default()
        blobhelper.download_blob(None, test_oid, test_tid, fn)
        self.assertTrue(os.path.exists(fn))

    def test_download_not_found(self):
        fn = os.path.join(self.blob_dir, '0001')
        blobhelper = self._make_default(download_action=None)
        blobhelper.download_blob(None, test_oid, test_tid, fn)
        self.assertFalse(os.path.exists(fn))

    def test_upload_without_tid(self):
        fn = os.path.join(self.blob_dir, '0001')
        blobhelper = self._make_default()
        blobhelper.upload_blob(None, test_oid, None, fn)
        self.assertEqual(self.uploaded, (1, None, fn))

    def test_upload_with_tid(self):
        fn = os.path.join(self.blob_dir, '0001')
        blobhelper = self._make_default()
        blobhelper.upload_blob(None, test_oid, test_tid, fn)
        self.assertEqual(self.uploaded, (1, 2, fn))

    def test_loadBlob_unshared_exists(self):
        blobhelper = self._make_default()
        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        os.makedirs(os.path.dirname(fn))
        write_file(fn, 'blob here')
        res = blobhelper.loadBlob(None, test_oid, test_tid)
        self.assertEqual(fn, res)

    def test_loadBlob_unshared_download(self):
        blobhelper = self._make_default()
        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        res = blobhelper.loadBlob(None, test_oid, test_tid)
        self.assertEqual(fn, res)

    def test_loadBlob_unshared_missing(self):
        blobhelper = self._make_default(download_action=None)
        from ZODB.POSException import POSKeyError
        self.assertRaises(POSKeyError, blobhelper.loadBlob, None, test_oid, test_tid)

    def test_openCommittedBlobFile_as_file(self):
        blobhelper = self._make_default()
        with blobhelper.openCommittedBlobFile(None, test_oid, test_tid) as f:
            if not PY3:
                self.assertEqual(f.__class__, file) # pylint:disable=undefined-variable
            self.assertEqual(f.read(), b'blob here')

    def test_openCommittedBlobFile_as_blobfile(self):
        blobhelper = self._make_default()
        from ZODB.blob import Blob
        from ZODB.blob import BlobFile
        b = Blob()
        with blobhelper.openCommittedBlobFile(None, test_oid, test_tid, blob=b) as f:
            self.assertEqual(f.__class__, BlobFile)
            self.assertEqual(f.read(), b'blob here')

    def test_openCommittedBlobFile_retry_as_file(self):
        loadBlob_calls = []
        blobhelper = self._make_default()
        orig_loadBlobInternal = blobhelper._loadBlobInternal
        def loadBlob_wrapper(cursor, oid, serial, blob_lock):
            fn = orig_loadBlobInternal(cursor, oid, serial, blob_lock)
            if not loadBlob_calls:
                os.remove(fn)
                loadBlob_calls.append(1)
            return fn

        blobhelper._loadBlobInternal = loadBlob_wrapper
        with blobhelper.openCommittedBlobFile(None, test_oid, test_tid) as f:
            self.assertEqual(loadBlob_calls, [1])
            if not PY3:
                self.assertEqual(f.__class__, file) # pylint:disable=undefined-variable
            self.assertEqual(f.read(), b'blob here')

    def test_openCommittedBlobFile_retry_as_blobfile(self):
        loadBlob_calls = []
        blobhelper = self._make_default()

        orig_loadBlobInternal = blobhelper._loadBlobInternal
        def loadBlob_wrapper(cursor, oid, serial, blob_lock):
            fn = orig_loadBlobInternal(cursor, oid, serial, blob_lock)
            if not loadBlob_calls:
                os.remove(fn)
                loadBlob_calls.append(1)
            return fn

        blobhelper._loadBlobInternal = loadBlob_wrapper
        from ZODB.blob import Blob
        from ZODB.blob import BlobFile
        b = Blob()
        with blobhelper.openCommittedBlobFile(None, test_oid, test_tid, b) as f:
            self.assertEqual(loadBlob_calls, [1])
            self.assertEqual(f.__class__, BlobFile)
            self.assertEqual(f.read(), b'blob here')

    def test_storeBlob_unshared(self):
        called = []
        dummy_txn = object()

        def store_func(oid, tid, data, txn):
            called.append((oid, tid, data, txn))

        fn = os.path.join(self.blob_dir, 'newblob')
        write_file(fn, 'here a blob')

        blobhelper = self._make_default()
        self.assertFalse(blobhelper.txn_has_blobs)
        blobhelper.storeBlob(None, store_func, test_oid, test_tid, 'blob pickle',
                             fn, '', dummy_txn)
        self.assertFalse(os.path.exists(fn))
        self.assertTrue(blobhelper.txn_has_blobs)
        self.assertEqual(called,
                         [(test_oid, test_tid, 'blob pickle', dummy_txn)])
        self.assertEqual(self.uploaded[:2], (1, None))
        target_fn = self.uploaded[2]
        self.assertEqual(read_file(target_fn), 'here a blob')

    def test_restoreBlob_unshared(self):
        fn = os.path.join(self.blob_dir, 'newblob')
        write_file(fn, 'here a blob')
        blobhelper = self._make_default()
        blobhelper.restoreBlob(None, test_oid, test_tid, fn)
        self.assertEqual(self.uploaded[:2], (1, 2))
        target_fn = self.uploaded[2]
        self.assertEqual(read_file(target_fn), 'here a blob')

    def test_copy_undone_unshared(self):
        blobhelper = self._make_default()
        blobhelper.copy_undone(None, None)
        self.assertFalse(blobhelper.txn_has_blobs)

    def test_after_pack_unshared(self):
        blobhelper = self._make_default()
        blobhelper.after_pack(None, None)  # No-op



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""MySQL adapter for RelStorage.

Connection parameters supported by MySQLdb:

host
    string, host to connect
user
    string, user to connect as
passwd
    string, password to use
db
    string, database to use
port
    integer, TCP/IP port to connect to
unix_socket
    string, location of unix_socket (UNIX-ish only)
conv
    mapping, maps MySQL FIELD_TYPE.* to Python functions which convert a
    string to the appropriate Python type
connect_timeout
    number of seconds to wait before the connection attempt fails.
compress
    if set, gzip compression is enabled
named_pipe
    if set, connect to server via named pipe (Windows only)
init_command
    command which is run once the connection is created
read_default_file
    see the MySQL documentation for mysql_options()
read_default_group
    see the MySQL documentation for mysql_options()
client_flag
    client flags from MySQLdb.constants.CLIENT
load_infile
    int, non-zero enables LOAD LOCAL INFILE, zero disables
"""
from __future__ import absolute_import
from __future__ import print_function

from relstorage.adapters.mysql.adapter import MySQLAdapter


assert MySQLAdapter



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Implementation of the high-level packing algorithm.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time

from persistent.timestamp import TimeStamp

from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64

from relstorage._compat import OID_SET_TYPE
from relstorage._compat import metricmethod
from .util import writable_storage_method

logger = __import__('logging').getLogger(__name__)

class Pack(object):

    __slots__ = (
        'options',
        'locker',
        'connmanager',
        'blobhelper',
        'cache',
        'packundo',
        'stats',
    )

    def __init__(self, options, adapter, blobhelper, cache):
        self.options = options
        self.locker = adapter.locker
        self.connmanager = adapter.connmanager
        self.packundo = adapter.packundo.with_options(options)
        self.stats = adapter.stats
        self.blobhelper = blobhelper
        self.cache = cache

    def __pre_pack(self, t, referencesf):
        logger.info("pack: beginning pre-pack")

        # In 2019, Unix timestamps look like
        #            1564006806.0
        # While 64-bit integer TIDs for the same timestamp look like
        #    275085010696509852
        #
        # Multiple TIDs can map to a single Unix timestamp.
        # For example, the 9 integers between 275085010624927044 and
        # 275085010624927035 all map to 1564006804.9999998.
        #
        # Therefore, Unix timestamps are ambiguous, especially if we're committing
        # multiple transactions rapidly (within the resolution of the underlying TID
        # clock).
        # This ambiguity mostly matters for unit tests, where we do commit rapidly.
        #
        # To help them out, we accept 64-bit integer TIDs to specify an exact
        # transaction to pack to.

        # We also allow None or a negative number to mean "current committed transaction".
        if t is None:
            t = -1

        if t > 275085010696509852:
            # Must be a TID.

            # This magic number generates 'Value too large to be stored in data type' when
            # given to time.gmtime() as used to originally generate TimeStamp values,
            # though it can be converted back. It also happens to specify July 2019
            # when this feature was implemented.

            # Turn it back into a time.time() for later logging
            ts = TimeStamp(int64_to_8bytes(t))
            logger.debug(
                "Treating requested pack time %s as TID meaning %s",
                t, ts
            )
            best_pack_tid_int = t
            t = ts.timeTime()
        elif t < 0 or t >= time.time():
            # Packing for the current time or in the future means to pack
            # to the lastest commit in the database. This matters if not all
            # machine clocks are synchronized.
            best_pack_tid_int = self.packundo.MAX_TID - 1
        else:
            # Find the latest commit before or at the pack time.
            # Note that several TIDs will fit in the resolution of a time.time(),
            # so this is slightly ambiguous.
            requested_pack_ts = TimeStamp(*time.gmtime(t)[:5] + (t % 60,))
            requested_pack_tid = requested_pack_ts.raw()
            requested_pack_tid_int = bytes8_to_int64(requested_pack_tid)

            best_pack_tid_int = requested_pack_tid_int

        tid_int = self.packundo.choose_pack_transaction(best_pack_tid_int)

        if tid_int is None:
            logger.debug("all transactions before %s have already "
                         "been packed", time.ctime(t))
            return

        s = time.ctime(TimeStamp(int64_to_8bytes(tid_int)).timeTime())
        logger.info("Analyzing transactions committed %s or before (TID %d)",
                    s, tid_int)

        # In pre_pack, the adapter fills tables with
        # information about what to pack.  The adapter
        # must not actually pack anything yet.
        def get_references(state):
            """Return an iterable of the set of OIDs the given state refers to."""
            if not state:
                return ()

            return {bytes8_to_int64(oid) for oid in referencesf(state)}

        self.packundo.pre_pack(tid_int, get_references)
        logger.info("pack: pre-pack complete")
        return tid_int

    def __pack_to(self, tid_int):
        # Now pack. We'll get a callback for every oid/tid removed,
        # and we'll use that to keep caches consistent.
        # In the common case of using zodbpack, this will rewrite the
        # persistent cache on the machine running zodbpack.
        oids_removed = OID_SET_TYPE()
        def invalidate_cached_data(
                oid_int, tid_int,
                cache=self.cache,
                blob_invalidate=self.blobhelper.after_pack,
                keep_history=self.options.keep_history,
                oids=oids_removed
        ):
            # pylint:disable=dangerous-default-value
            # Flush the data from the local/global cache. It's quite likely that
            # a fair amount if it is now useless. We almost certainly want to
            # establish new checkpoints. The alternative is to clear out
            # data that we know we removed in the pack; we *do* keep track
            # of that, it's what's passed to the `packed_func`, so we could probably
            # piggyback on that in some fashion.
            #
            # Having consistent cache data became especially important
            # when loadSerial() began using the cache: Since that function
            # is allowed to return non-transactional data, it wouldn't be
            # great for it to return data that is no longer in the DB,
            # only the cache. I *think* this is only an issue in the tests,
            # that use the storage in a non-conventional way after packing,
            # by directly verifying that loadSerial() doesn't return data,
            # when in a real use we could only get there through detecting a conflict
            # in the database at commit time, with locks involved.
            cache.remove_cached_data(oid_int, tid_int)
            # Clean up blob files. This currently does nothing
            # if we're a blob cache, but it could.
            blob_invalidate(oid_int, tid_int)
            # If we're not keeping history, we need to remove all the cached
            # data for a particular OID, no matter what key it was under:
            # there was only one way to access it.
            if not keep_history:
                oids.add(oid_int)

        self.packundo.pack(tid_int,
                           packed_func=invalidate_cached_data)
        self.cache.remove_all_cached_data_for_oids(oids_removed)

    @writable_storage_method
    @metricmethod
    def pack(self, t, referencesf, prepack_only=False, skip_prepack=False):
        """Pack the storage. Holds the pack lock for the duration."""
        # pylint:disable=too-many-branches,unused-argument

        prepack_only = prepack_only or self.options.pack_prepack_only
        skip_prepack = skip_prepack or self.options.pack_skip_prepack

        if prepack_only and skip_prepack:
            raise ValueError('Pick either prepack_only or skip_prepack.')

        lock_conn, lock_cursor = self.connmanager.open_for_pack_lock()
        try:
            self.locker.hold_pack_lock(lock_cursor)
            try:
                if not skip_prepack:
                    tid_int = self.__pre_pack(t, referencesf)
                else:
                    # Need to determine the tid_int from the pack_object table
                    tid_int = self.packundo._find_pack_tid()

                if not prepack_only:
                    self.__pack_to(tid_int)
            finally:
                self.locker.release_pack_lock(lock_cursor)
        finally:
            self.connmanager.rollback_and_close(lock_conn, lock_cursor)

        self.stats.large_database_change()



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""MySQL adapter for RelStorage.

Connection parameters supported by MySQLdb:

host
    string, host to connect
user
    string, user to connect as
passwd
    string, password to use
db
    string, database to use
port
    integer, TCP/IP port to connect to
unix_socket
    string, location of unix_socket (UNIX-ish only)
conv
    mapping, maps MySQL FIELD_TYPE.* to Python functions which convert a
    string to the appropriate Python type
connect_timeout
    number of seconds to wait before the connection attempt fails.
compress
    if set, gzip compression is enabled
named_pipe
    if set, connect to server via named pipe (Windows only)
init_command
    command which is run once the connection is created
read_default_file
    see the MySQL documentation for mysql_options()
read_default_group
    see the MySQL documentation for mysql_options()
client_flag
    client flags from MySQLdb.constants.CLIENT
load_infile
    int, non-zero enables LOAD LOCAL INFILE, zero disables
"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import json

from zope.interface import implementer

from relstorage._compat import iteritems
from relstorage._compat import metricmethod_sampled
from relstorage._util import TRACE

from ..adapter import AbstractAdapter

from ..dbiter import HistoryFreeDatabaseIterator
from ..dbiter import HistoryPreservingDatabaseIterator
from ..interfaces import IRelStorageAdapter
from ..interfaces import UnableToLockRowsToReadCurrentError
from ..interfaces import UnableToLockRowsToModifyError
from ..interfaces import UnableToAcquireCommitLockError
from ..poller import Poller
from ..scriptrunner import ScriptRunner
from ..batch import RowBatcher

from . import drivers
from .connmanager import MySQLdbConnectionManager
from .locker import MySQLLocker
from .mover import MySQLObjectMover

from .oidallocator import MySQLOIDAllocator
from .packundo import MySQLHistoryFreePackUndo
from .packundo import MySQLHistoryPreservingPackUndo
from .schema import MySQLSchemaInstaller
from .schema import MySQLVersionDetector
from .stats import MySQLStats
from .txncontrol import MySQLTransactionControl

logger = logging.getLogger(__name__)


@implementer(IRelStorageAdapter)
class MySQLAdapter(AbstractAdapter):
    """MySQL adapter for RelStorage."""
    # pylint:disable=too-many-instance-attributes

    driver_options = drivers

    def __init__(self, options=None, oidallocator=None,
                 version_detector=None,
                 connmanager=None,
                 locker=None,
                 mover=None,
                 **params):
        self._params = params
        self.oidallocator = oidallocator
        self.version_detector = version_detector
        self.connmanager = connmanager
        self.locker = locker
        self.mover = mover
        super(MySQLAdapter, self).__init__(options)

    def _create(self):
        options = self.options
        driver = self.driver
        params = self._params


        if self.version_detector is None:
            self.version_detector = MySQLVersionDetector()

        if self.connmanager is None:
            self.connmanager = MySQLdbConnectionManager(
                driver,
                params=params,
                options=options,
            )
        if self.locker is None:
            self.locker = MySQLLocker(
                options=options,
                driver=driver,
                batcher_factory=RowBatcher,
                version_detector=self.version_detector,
            )
        self.runner = ScriptRunner()
        self.schema = MySQLSchemaInstaller(
            driver=driver,
            connmanager=self.connmanager,
            runner=self.runner,
            keep_history=self.keep_history,
            version_detector=self.version_detector,
        )
        if self.mover is None:
            self.mover = MySQLObjectMover(
                driver,
                options=options,
            )

        if self.oidallocator is None:
            self.oidallocator = MySQLOIDAllocator(driver)

        self.poller = Poller(
            self.driver,
            keep_history=self.keep_history,
            runner=self.runner,
            revert_when_stale=options.revert_when_stale,
            transactions_may_go_backwards=(
                self.connmanager.replica_selector is not None
                or self.connmanager.ro_replica_selector is not None
            )
        )

        self.txncontrol = MySQLTransactionControl(
            connmanager=self.connmanager,
            poller=self.poller,
            keep_history=self.keep_history,
            Binary=driver.Binary,
        )

        if self.keep_history:
            self.packundo = MySQLHistoryPreservingPackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            self.dbiter = HistoryPreservingDatabaseIterator(
                driver,
            )
        else:
            self.packundo = MySQLHistoryFreePackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            self.dbiter = HistoryFreeDatabaseIterator(
                driver,
            )

        self.stats = MySQLStats(
            connmanager=self.connmanager,
            keep_history=self.keep_history
        )

    def new_instance(self):
        return type(self)(
            options=self.options,
            oidallocator=self.oidallocator.new_instance(),
            version_detector=self.version_detector,
            connmanager=self.connmanager,
            locker=self.locker,
            mover=self.mover,
            **self._params
        )

    def __str__(self):
        parts = [self.__class__.__name__]
        if self.keep_history:
            parts.append('history preserving')
        else:
            parts.append('history free')
        p = self._params.copy()
        if 'passwd' in p:
            del p['passwd']
        p = sorted(iteritems(p))
        parts.extend('%s=%r' % item for item in p)
        return ", ".join(parts)

    # A temporary magic variable as we move TID allocation into some
    # databases; with an external clock, we *do* need to sleep waiting for
    # TIDs to change in a manner we can exploit; that or we need to be very
    # careful about choosing pack times.
    RS_TEST_TXN_PACK_NEEDS_SLEEP = 1

    @metricmethod_sampled
    def lock_database_and_choose_next_tid(self,
                                          cursor,
                                          username,
                                          description,
                                          extension):
        proc = 'lock_and_choose_tid'
        args = ()
        if self.keep_history:
            # (packed, username, descr, extension)
            proc = proc + '(%s, %s, %s, %s)'
            args = (False, username, description, extension)

        multi_results = self.driver.callproc_multi_result(cursor, proc, args)
        tid, = multi_results[0][0]
        logger.log(TRACE, "Locked database only to choose tid %s", tid)
        return tid

    @metricmethod_sampled
    def lock_database_and_move(self,
                               store_connection, load_connection,
                               transaction_has_blobs,
                               ude,
                               commit=True,
                               committing_tid_int=None,
                               after_selecting_tid=lambda tid: None):
        if not self.version_detector.supports_good_stored_procs(store_connection.cursor):
            # XXX: When can we drop this? Probably not until AppVeyor upgrades
            # MySQL past 5.7.12.
            return super(MySQLAdapter, self).lock_database_and_move(
                store_connection, load_connection,
                transaction_has_blobs,
                ude,
                commit=commit,
                committing_tid_int=committing_tid_int,
                after_selecting_tid=after_selecting_tid)

        params = (committing_tid_int, commit)
        # (p_committing_tid, p_commit)
        proc = 'lock_and_choose_tid_and_move(%s, %s)'
        if self.keep_history:
            params += ude
            # (p_committing_tid, p_commit, p_user, p_desc, p_ext)
            proc = 'lock_and_choose_tid_and_move(%s, %s, %s, %s, %s)'

        try:
            multi_results = self.driver.callproc_multi_result(
                store_connection.cursor,
                proc,
                params,
                exit_critical_phase=commit
            )
        except self.driver.lock_exceptions:
            self.locker.reraise_commit_lock_error(
                store_connection.cursor,
                proc,
                UnableToAcquireCommitLockError,
            )

        tid_int, = multi_results[0][0]
        after_selecting_tid(tid_int)
        logger.log(TRACE, "Locked database and moved rows for tid %s", tid_int)
        return tid_int, "-"

    DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = False

    def _best_lock_objects_and_detect_conflicts(self, cursor, read_current_oids):
        read_current_param = None
        if read_current_oids:
            # In MySQL 8, we could pass in a JSON array and use JSON_TABLE
            # to join directly against the data.

            # In earlier versions, we could do some tricks with strings and
            # preparing dynamic SQL.

            # In all versions, we could write a query like
            #   select tid from object_state where json_contains('[1, 8]', cast(zoid as json), '$')
            # but that is very slow (entire table is scanned).
            #
            # We pass the string array, parse it as json, loop over it to put in a temp table
            # and join that.
            #
            # I also benchmarked sending the data in a format suitable
            # for concatenating to the end of 'INSERT ... VALUES' and
            # using dynamic SQL in the stored proc to execute that
            # ('PREPARE stmt FROM @str; EXECUTE stmt') and it
            # benchmarked essentially the same. Usually there are a
            # small number of things being readCurrent so it probably
            # doesn't matter.
            read_current_param = json.dumps(list(read_current_oids.items()))

        proc = 'lock_objects_and_detect_conflicts(%s, %s)'
        try:
            multi_results = self.driver.callproc_multi_result(
                cursor,
                proc,
                (read_current_param, 0)
            )
        except self.locker.lock_exceptions as e:
            # On MySQL 5.7, the time-based mechanism to determine that
            # we failed to take NOWAIT shared locks is not reliable
            # when the commit lock timeout is very small (close to 1),
            # because it can take several seconds for the procedure te
            # error out. Thus we provide a specific error message that
            # we key off. (We don't change errno or state or anything
            # like that in case anybody introspects that stuff.)
            if 'shared locks' in str(e):
                self.locker.reraise_commit_lock_error(
                    cursor,
                    self._describe_best_lock_objects_and_detect_conflicts(),
                    UnableToLockRowsToReadCurrentError
                )
            elif 'exclusive locks' in str(e):
                self.locker.reraise_commit_lock_error(
                    cursor,
                    self._describe_best_lock_objects_and_detect_conflicts(),
                    UnableToLockRowsToModifyError
                )
            raise

        # There's always a useless last result, the result of the stored procedure itself.
        proc_result = multi_results.pop()
        assert not proc_result, proc_result
        assert len(multi_results) == 1, multi_results
        conflicts = multi_results[0]

        return conflicts

    def _describe_best_lock_objects_and_detect_conflicts(self):
        return 'lock_objects_and_detect_conflicts(%s)'



"""
Connection support for Blobs tests
"""
from __future__ import absolute_import
from __future__ import print_function

from ZODB.interfaces import IBlob
from ZODB.blob import Blob
import transaction

from . import TestBlobMixin

class TestConnectionBlobMixin(TestBlobMixin):

    first_data = b"(1) I'm a happy Blob."

    def testPuttingInConnection(self):
        # Putting a Blob into a Connection works like every other object:
       # Connections handle Blobs specially. To demonstrate that, we
        # first need a Blob with some data:

        self.blob = blob = Blob()
        with blob.open("w") as f:
            f.write(self.first_data)

        connection = self.database.open()
        root = connection.root()
        root['myblob'] = blob
        transaction.commit()
        connection.close()

    def testCommitWithoutOpen(self):
        # We can also commit a transaction that seats a blob into place without:
        # calling the blob's open method:
        transaction.begin()
        connection = self.database.open()
        anotherblob = Blob()
        root = connection.root()
        root['anotherblob'] = anotherblob
        transaction.commit()
        connection.close()

    def testGettingStuffOut(self):
        # Getting stuff out of there works similarly:
        self.testPuttingInConnection()

        transaction2 = transaction.TransactionManager()
        connection2 = self.database.open(transaction_manager=transaction2)
        root = connection2.root()
        blob2 = root['myblob']
        self.assertTrue(IBlob.providedBy(blob2))

        with blob2.open("r") as f:
            data = f.read()
        self.assertEqual(data, self.first_data)
        transaction2.abort()
        connection2.close()

    def testMVCC(self):
        # MVCC also works.
        self.testPuttingInConnection()

        connection = self.database.open()
        transaction3 = transaction.TransactionManager()
        connection3 = self.database.open(transaction_manager=transaction3)

        second_data = b"(2) I am an ecstatic Blob."
        with connection.root()['myblob'].open('w') as f:
            f.write(second_data)
        transaction.commit()

        with connection3.root()['myblob'].open('r') as f:
            data = f.read()
        self.assertEqual(data, self.first_data)

        transaction3.abort()
        connection.close()
        connection3.close()



"""relstorage.tests package"""

import abc
import contextlib
import os
import unittest
import functools

import transaction

from ZODB.Connection import Connection
from ZODB.tests.util import clear_transaction_syncs

from relstorage._compat import ABC
from relstorage.options import Options
from relstorage.adapters.sql import DefaultDialect

try:
    from unittest import mock as _mock
except ImportError: # Python 2
    import mock as _mock

mock = _mock

@contextlib.contextmanager
def _fakeSubTest(*_args, **_kwargs):
    yield

class TestCase(unittest.TestCase):
    """
    General tests that may use databases, connections and
    transactions, but don't have any specific requirements or
    framework to do so.

    This class supplies some supporting help for assertions and
    cleanups.
    """
    # Avoid deprecation warnings; 2.7 doesn't have
    # assertRaisesRegex or assertRegex
    assertRaisesRegex = getattr(
        unittest.TestCase,
        'assertRaisesRegex',
        None
    ) or getattr(unittest.TestCase, 'assertRaisesRegexp')
    assertRegex = getattr(
        unittest.TestCase,
        'assertRegex',
        None
    ) or getattr(unittest.TestCase, 'assertRegexpMatches')

    # 2.7 doesn't have subtest.
    subTest = getattr(
        unittest.TestCase,
        'subTest',
        None
    ) or _fakeSubTest

    none = unittest.TestCase.assertIsNone

    def setUp(self):
        super(TestCase, self).setUp()
        # This is done by ZODB.tests.util.TestCase, but
        # not stored anywhere.
        # XXX: As of ZODB 5.5.1, that class also doesn't handle
        # Python 3 correctly either. File a bug about this.
        name = self.__class__.__name__

        # Python 2
        mname = getattr(self, '_TestCase__testMethodName', '')
        if not mname:
            # Python 3
            mname = getattr(self, '_testMethodName', '')
        if mname:
            name += '-' + mname
        self.rs_temp_prefix = name
        self.addCleanup(clear_transaction_syncs)

    def __close_connection(self, connection):
        if connection.opened:
            try:
                connection.close()
            except KeyError:
                # From the transaction manager's list of syncs.
                # This can happen if clear_transaction_syncs()
                # has already been called.
                pass

            connection.close = lambda *_, **__: None

    def __close(self, thing):
        try:
            thing.close()
        except KeyError:
            # As for Connection; a DB closes all its connections.
            pass

    def _closing(self, o):
        """
        Close the object using its 'close' method *after* invoking
        all of the `tearDown` stack, and even running if `setUp`
        fails.

        This is just a convenience wrapper around `addCleanup`. This
        exists to make it easier to call.

        Failures (exceptions) in your cleanup function will still
        result in the test failing, but all registered cleanups will
        still be run.

        Returns the given object.
        """
        __traceback_info__ = o
        # Don't capture o.close now, it could get swizzled later to prevent it
        # from doing things. For example, DB.close() sets self.close to a no-op

        if isinstance(o, Connection):
            meth = self.__close_connection
        else:
            meth = self.__close
        self.addCleanup(meth, o)
        return o

    def tearDown(self):
        try:
            transaction.abort()
        finally:
            super(TestCase, self).tearDown()

    def assertIsEmpty(self, container, msg=None):
        self.assertLength(container, 0, msg)

    assertEmpty = assertIsEmpty

    def assertLength(self, container, length, msg=None):
        self.assertEqual(len(container), length,
                         '%s -- %s' % (msg, container) if msg else container)

def skipIfNoConcurrentWriters(func):
    # Skip the test if only one writer is allowed at a time.
    # e.g., sqlite
    @functools.wraps(func)
    def f(self):
        if self._storage._adapter.WRITING_REQUIRES_EXCLUSIVE_LOCK:
            self.skipTest("Storage doesn't support multiple writers")
        func(self)
    return f


class StorageCreatingMixin(ABC):

    keep_history = None # Override
    driver_name = None # Override.
    zap_slow = False # Override

    @abc.abstractmethod
    def make_adapter(self, options):
        # abstract method
        raise NotImplementedError

    @abc.abstractmethod
    def get_adapter_class(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_adapter_zconfig(self):
        """
        Return the part of the ZConfig string that makes the adapter.

        That is, return the <postgresql>, <mysql> or <oracle> section.

        Return text (unicode).
        """
        raise NotImplementedError

    def get_adapter_zconfig_replica_conf(self):
        return os.path.join(os.path.dirname(__file__), 'replicas.conf')

    @abc.abstractmethod
    def verify_adapter_from_zconfig(self, adapter):
        """
        Assert that the adapter configured from get_adapter_zconfig
        is properly configured.
        """
        raise NotImplementedError

    def _wrap_storage(self, storage):
        return storage

    DEFAULT_COMMIT_LOCK_TIMEOUT = 10

    def make_storage(self, zap=True, **kw):
        from . import util
        from relstorage.storage import RelStorage

        if ('cache_servers' not in kw
                and 'cache_module_name' not in kw
                and kw.get('share_local_cache', True)):
            if util.CACHE_SERVERS and util.CACHE_MODULE_NAME:
                kw['cache_servers'] = util.CACHE_SERVERS
                kw['cache_module_name'] = util.CACHE_MODULE_NAME
        if 'cache_prefix' not in kw:
            kw['cache_prefix'] = type(self).__name__ + self._testMethodName
        if 'cache_local_dir' not in kw:
            # Always use a persistent cache. This helps discover errors in
            # the persistent cache.
            # These tests run in a temporary directory that gets cleaned up, so the CWD is
            # appropriate. BUT: it should be an abspath just in case we change directories
            kw['cache_local_dir'] = os.path.abspath('.')
        if 'commit_lock_timeout' not in kw:
            # Cut this way down so we get better feedback.
            kw['commit_lock_timeout'] = self.DEFAULT_COMMIT_LOCK_TIMEOUT

        assert self.driver_name
        options = Options(keep_history=self.keep_history, driver=self.driver_name, **kw)
        adapter = self.make_adapter(options)
        storage = RelStorage(adapter, options=options)
        if zap:
            storage.zap_all(slow=self.zap_slow)
        return self._wrap_storage(storage)

class MockConnection(object):
    rolled_back = False
    closed = False
    replica = None
    committed = False

    def commit(self):
        self.committed = True

    def rollback(self):
        self.rolled_back = True

    def close(self):
        self.closed = True

    def cursor(self):
        return MockCursor(self)

class MockCursor(object):
    closed = False
    sort_sequence_params = False

    def __init__(self, conn=None):
        self.executed = []
        self.inputsizes = {}
        self.results = []
        self.many_results = None
        self.connection = conn

    def setinputsizes(self, **kw):
        self.inputsizes.update(kw)

    def execute(self, stmt, params=None):
        if self.sort_sequence_params and isinstance(params, (tuple, list)):
            params = sorted(sorted(p) if isinstance(p, (tuple, list)) else p
                            for p in params)

        params = tuple(params) if isinstance(params, list) else params
        self.executed.append((stmt, params))

    def fetchone(self):
        return self.results.pop(0)

    def fetchall(self):
        if self.many_results:
            return self.many_results.pop(0)
        r = self.results
        self.results = None
        return r

    def close(self):
        self.closed = True

    def __iter__(self):
        for row in self.results:
            yield row

class MockOptions(Options):
    cache_module_name = '' # disable
    cache_servers = ''
    cache_local_mb = 1
    cache_local_dir_count = 1 # shrink

    @classmethod
    def from_args(cls, **kwargs):
        inst = cls()
        for k, v in kwargs.items():
            setattr(inst, k, v)
        return inst

    def __setattr__(self, name, value):
        if name not in Options.valid_option_names():
            raise AttributeError("Invalid option", name) # pragma: no cover
        object.__setattr__(self, name, value)

class MockConnectionManager(object):

    isolation_load = 'SERIALIZABLE'

    def __init__(self, driver=None):
        if driver is None:
            self.driver = MockDriver()

    def rollback_quietly(self, conn, cursor): # pylint:disable=unused-argument
        if hasattr(conn, 'rollback'):
            conn.rollback()

    rollback_store_quietly = rollback_quietly

    def rollback_and_close(self, conn, cursor):
        self.rollback_quietly(conn, cursor)
        if conn:
            conn.close()
        if cursor:
            cursor.close()

    def open_for_load(self):
        conn = MockConnection()
        return conn, self.configure_cursor(conn.cursor())

    def configure_cursor(self, cur):
        return cur

    open_for_store = open_for_load

    def restart_load(self, conn, cursor, needs_rollback=True):
        pass

    restart_store = restart_load

    def cursor_for_connection(self, conn):
        return conn.cursor()

    def open_and_call(self, func):
        return func(*self.open_for_load())

class MockPackUndo(object):
    pass

class MockOIDAllocator(object):
    pass

class MockQuery(object):

    def __init__(self, raw):
        self.raw = raw

    def execute(self, cursor, params=None):
        cursor.execute(self.raw, params)

class MockPoller(object):

    poll_query = MockQuery('SELECT MAX(tid) FROM object_state')
    last_requested_range = None
    def __init__(self, driver=None):
        self.driver = driver or MockDriver()
        self.changes = []  # [(oid, tid)]
        self.poll_tid = 1
        self.poll_changes = None

    def poll_invalidations(self, _conn, _cursor, _since):
        return self.poll_changes, self.poll_tid

    def get_current_tid(self, _cursor):
        return self.poll_tid


class DisconnectedException(Exception):
    pass

class CloseException(Exception):
    pass

class LockException(Exception):
    pass

class MockDriver(object):
    Binary = bytes
    supports_64bit_unsigned_id = True
    cursor_arraysize = 64
    disconnected_exceptions = (DisconnectedException,)
    close_exceptions = (CloseException,)
    lock_exceptions = (LockException,)
    illegal_operation_exceptions = ()

    dialect = DefaultDialect()

    isolation_load = 'SERIALIZABLE'
    def connection_may_need_rollback(self, conn): # pylint:disable=unused-argument
        return True
    connection_may_need_commit = connection_may_need_rollback

    def enter_critical_phase_until_transaction_end(self, connection, cursor):
        "Does nothing"

    is_in_critical_phase = enter_critical_phase_until_transaction_end
    exit_critical_phase = enter_critical_phase_until_transaction_end

    def commit(self, conn):
        conn.commit()

    def rollback(self, conn):
        conn.rollback()

    def synchronize_cursor_for_rollback(self, cursor):
        cursor.fetchall()

class MockObjectMover(object):
    def __init__(self):
        self.data = {}  # {oid_int: (state, tid_int)}

    def load_current(self, _cursor, oid_int):
        return self.data.get(oid_int, (None, None))

    def current_object_tids(self, _cursor, oids, timeout=None):
        # pylint:disable=unused-argument
        return {
            oid: self.data[oid][1]
            for oid in oids
            if oid in self.data
        }

class MockLocker(object):

    def release_commit_lock(self, cursor):
        pass

class MockAdapter(object):

    def __init__(self):
        self.driver = MockDriver()
        self.connmanager = MockConnectionManager()
        self.packundo = MockPackUndo()
        self.oidallocator = MockOIDAllocator()
        self.poller = MockPoller(self.driver)
        self.mover = MockObjectMover()
        self.locker = MockLocker()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for dialects.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase

from .. import dialect

class TestMissingDialect(TestCase):

    def test_boolean(self):
        d = dialect._MissingDialect()
        self.assertFalse(d)


class TestCompiler(TestCase):

    def test_prepare_no_datatypes(self):

        class C(dialect.Compiler):
            def _next_prepared_stmt_name(self, query):
                return 'my_stmt'

            def _find_datatypes_for_prepared_query(self):
                return ()

        compiler = C(None, dialect.DefaultDialect())

        stmt, execute, convert = compiler.prepare()

        self.assertEqual(
            stmt,
            'PREPARE my_stmt AS '
        )

        self.assertEqual(
            execute,
            'EXECUTE my_stmt'
        )

        # We get the default dictionary converter even if we
        # don't need it.

        self.assertEqual(
            [],
            convert({'a': 42})
        )

    def test_prepare_named_datatypes(self):

        compiler = dialect.Compiler(None, dialect.DefaultDialect())
        compiler.placeholders[object()] = 'name'

        _s, _x, convert = compiler.prepare()

        self.assertEqual(
            [64],
            convert({'name': 64})
        )


class TestDialectAware(TestCase):

    def test_bind_none(self):

        aware = dialect.DialectAware()

        with self.assertRaisesRegex(TypeError, 'no dialect found'):
            aware.bind(None)

    def test_bind_dialect(self):
        class Dialect(dialect.DefaultDialect):
            def bind(self, context):
                raise AssertionError("Not supposed to re-bind")

        d = Dialect()

        aware = dialect.DialectAware()
        new_aware = aware.bind(d)
        self.assertIsNot(aware, new_aware)
        self.assertIs(new_aware.dialect, d)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
from functools import partial

from relstorage.tests import TestCase

from . import LocalClient
from . import MockOptions
from . import list_lrukeys as list_lrukeys_
from . import list_lrufreq as list_lrufreq_
from .test_memcache_client import AbstractStateCacheTests


class LocalClientStrKeysValuesGenerationalTests(TestCase):
    """
    Depend on having an IGenerationalLRUCache as the backing storage.
    """

    class LocalClientStrKeysValues(LocalClient):
        # Make the cache accept and return str keys and values,
        # for ease of dealing with size limits,
        # bust mostly for compatibility with old tests, before LocalClient
        # implemented IStateCache

        def __setitem__(self, key, val):
            oid = int(key[0])
            tid = int(key[1])
            LocalClient.__setitem__(self, (oid, tid), (val, tid))

        def __getitem__(self, key):
            v = LocalClient.__getitem__(self, (int(key[0]), int(key[1])))
            if v is not None:
                v = v[0]
            return v

    def getClass(self):
        return self.LocalClientStrKeysValues

    Options = MockOptions

    def _makeOne(self, **kw):
        options = self.Options.from_args(**kw)
        inst = self.getClass()(options, 'pfx')
        inst.restore()
        return inst

    def test_bucket_sizes_without_compression(self):
        # pylint:disable=too-many-statements
        # LocalClient is a simple w-TinyLRU cache.  Confirm it keeps the right keys.
        c = self._makeOne(cache_local_compression='none')

        # The initial weight is the size of the cache object and the generations;
        # on a 64-bit platform, I get 224
        initial_weight = c._cache.weight
        c['55'] = b''
        # The weight of an empty entry is the overhead. On a64-bit platform
        # I get 96
        entry_weight = c._cache[5].weight

        # This limit will result in
        # eden and probation of 5, protected of 40. This means that eden
        # and probation each can hold one item, while protected can hold 4,
        # so our max size will be 60
        c.limit = (entry_weight + 10) * 5 + 1
        setattr(c, '_cache', True) # Avoid type inference warnings
        c.flush_all()

        list_lrukeys = partial(list_lrukeys_, c._cache)
        list_lrufreq = partial(list_lrufreq_, c._cache)
        def all_lrukeys():
            return {k: list_lrukeys(k) for k in ('eden', 'probation', 'protected')}
        def expected_weight(entry_count, python_weight):
            return python_weight + (entry_count * entry_weight) + initial_weight

        k = None

        for i in range(5):
            # add 10 bytes
            k = '%d%d' % (i, i)
            # This will go to eden, replacing any value that was there
            # into probation.
            c[k] = b'0123456789'


        # While we have the room, we initially put items into the protected
        # space when they graduate from eden.
        self.assertEqual(
            all_lrukeys(),
            {'eden': [4], 'probation': [], 'protected': [3, 2, 1, 0]}
        )
        self.assertEqual(c._cache.weight, expected_weight(5, 50))

        c['55'] = b'0123456789'

        # Right now, we're one entry over size, because we put k5
        # in eden, which dropped k4 to probation; since probation was empty, we
        # allowed it to stay there
        self.assertEqual(list_lrukeys('eden'), [5])
        self.assertEqual(list_lrukeys('probation'), [4])
        self.assertEqual(list_lrukeys('protected'), [3, 2, 1, 0])
        self.assertEqual(c._cache.weight, expected_weight(6, 60))

        v = c['22']
        self.assertEqual(v, b'0123456789')
        self.assertEqual(c._cache.weight, expected_weight(6, 60))

        del c[(1, 1)]
        c['11'] = b'b'
        self.assertEqual(list_lrukeys('eden'), [1])
        self.assertEqual(list_lrukeys('probation'), [5])
        self.assertEqual(list_lrukeys('protected'), [2, 3, 0])

        self.assertEqual(c._cache.weight, expected_weight(5, 41))

        for i in range(4):
            # add 10 bytes
            c[(-i, i)] = b'0123456789'
            # Notice that we're not promoting these through the layers. So
            # when we're done, we'll wind up with one key each in
            # eden and probation, and all the K keys in protected (since
            # they have been promoted)


        # x0 and x1 started in eden and got promoted to the probation ring,
        # from whence they were ejected because of never being accessed.
        # k2 was allowed to remain because it'd been accessed
        # more often
        self.assertEqual(list_lrukeys('eden'), [-3])
        self.assertEqual(list_lrukeys('probation'), [-2])
        self.assertEqual(list_lrukeys('protected'), [0, 2, 3])
        self.assertEqual(c._cache.weight, expected_weight(5, 50))

        #pprint.pprint(c._cache.stats())
        self.assertEqual(c[(-1, 0)], None)
        self.assertEqual(c[(-1, 1)], None)
        self.assertEqual(c[(-2, 2)], b'0123456789')
        self.assertEqual(c[(-3, 3)], b'0123456789')
        self.assertEqual(c[(2, 2)], b'0123456789')

        # Note that this last set of checks perturbed protected and probation;
        # We lost a key
        #pprint.pprint(c._cache.stats())
        self.assertEqual(list_lrukeys('eden'), [-3])
        self.assertEqual(list_lrukeys('probation'), [])
        self.assertEqual(list_lrukeys('protected'), [2, -2, 0, 3])


        self.assertEqual(c['00'], b'0123456789')
        self.assertEqual(c['00'], b'0123456789') # One more to increase its freq count
        self.assertEqual(c['11'], None)
        self.assertEqual(c['22'], b'0123456789')
        self.assertEqual(c['33'], b'0123456789')
        self.assertEqual(c['44'], None)
        self.assertEqual(c['55'], None)

        # Let's promote from probation, causing places to switch.
        # First, verify our current state after those gets.
        self.assertEqual(list_lrukeys('eden'), [-3])
        self.assertEqual(list_lrukeys('probation'), [])
        self.assertEqual(list_lrukeys('protected'), [3, 2, 0, -2])
        # Now get and switch
        c.__getitem__((-2, 2))
        self.assertEqual(list_lrukeys('eden'), [-3])
        self.assertEqual(list_lrukeys('probation'), [])
        self.assertEqual(list_lrukeys('protected'), [-2, 3, 2, 0])


        # Confirm frequency counts
        self.assertEqual(list_lrufreq('eden'), [2])
        self.assertEqual(list_lrufreq('probation'), [])
        self.assertEqual(list_lrufreq('protected'), [3, 2, 4, 4])
        # A brand new key is in eden, shifting eden to probation

        c[(100, 0)] = b'0123456789'

        # Now, because we had accessed k0 (probation) more than we'd
        # accessed the last key from eden (x3), that's the one we keep
        self.assertEqual(list_lrukeys('eden'), [100])
        self.assertEqual(list_lrukeys('probation'), [-3])
        self.assertEqual(list_lrukeys('protected'), [-2, 3, 2, 0])

        self.assertEqual(list_lrufreq('eden'), [1])
        self.assertEqual(list_lrufreq('probation'), [2])
        self.assertEqual(list_lrufreq('protected'), [3, 2, 4, 4])

        self.assertEqual(c._cache.weight, expected_weight(6, 60))

        self.assertEqual(c[(-3, 3)], b'0123456789')
        self.assertEqual(list_lrukeys('probation'), [])


class LocalClientOIDTests(AbstractStateCacheTests):
    # Uses true oid/int keys and state/tid values.

    Options = MockOptions

    tid = 34567
    key_tid = tid
    oid = 12345

    key = (oid, key_tid)
    value = (b'statebytes', tid)

    missing_key = (0, 1)

    def getClass(self):
        return LocalClient

    def test_cache_corruption_on_store(self):
        from relstorage.cache.interfaces import CacheConsistencyError
        c = self._makeOne(cache_local_dir=':memory:')
        c[self.key] = self.value
        # Different key, same real tid, different state. Uh-oh!
        with self.assertRaises(CacheConsistencyError):
            c[(self.oid, self.tid)] = (b'bad bytes', self.tid)

    def test_ctor(self):
        c = self._makeOne()
        self.assertEqual(c.limit, 1000000)
        self.assertEqual(c._value_limit, 16384)
        # cover
        self.assertIn('hits', c.stats())
        c.reset_stats()
        c.close()

        self.assertRaises(ValueError,
                          self._makeOne,
                          cache_local_compression='unsup')

    def test_set_and_get_string_compressed(self):
        c = self._makeOne(cache_local_compression='zlib')
        c[self.key] = self.value
        self.assertEqual(c[self.key], self.value)
        self.assertEqual(c[self.missing_key], None)

    def test_set_and_get_string_uncompressed(self):
        c = self._makeOne(cache_local_compression='none')
        c[self.key] = self.value
        self.assertEqual(c[self.key], self.value)
        self.assertEqual(c[self.missing_key], None)

    def test_set_and_get_object_too_large(self):
        c = self._makeOne(cache_local_compression='none')
        c[self.key] = (b'abcdefgh' * 10000, self.key_tid)
        self.assertEqual(c[self.key], None)

    def test_set_with_zero_space(self):
        c = self._makeOne(cache_local_mb=0)
        self.assertEqual(c.limit, 0)
        self.assertEqual(c._value_limit, 16384)
        c[self.key] = self.value
        c[self.missing_key] = (b'', 1)
        self.assertEqual(c[self.key], None)
        self.assertEqual(c[self.missing_key], None)

    def test_load_and_save(self):
        # pylint:disable=too-many-statements,too-many-locals
        import tempfile
        import shutil
        import os

        root_temp_dir = tempfile.mkdtemp(".rstest_cache")
        self.addCleanup(shutil.rmtree, root_temp_dir, True)
        # Intermediate directories will be auto-created
        temp_dir = os.path.join(root_temp_dir, 'child1', 'child2')

        c = self._makeOne(cache_local_dir=temp_dir)
        c.restore()
        # Doing the restore created the database.
        files = os.listdir(temp_dir)
        __traceback_info__ = files
        # There may be up to 3 files here, including the -wal and -shm
        # files, depending on how the database is closed and how the database
        # was configured. It also depends os when we look: some of the wal cleanup
        # we push to a background thread.
        def get_cache_files():
            return [x for x in os.listdir(temp_dir) if x.endswith('sqlite3')]
        cache_files = get_cache_files()
        len_initial_cache_files = len(cache_files)
        self.assertEqual(len_initial_cache_files, 1)
        # Saving an empty bucket does nothing
        self.assertFalse(c.save())

        key = (0, 1)
        val = (b'abc', 1)
        c[key] = val
        c.__getitem__(key) # Increment the count so it gets saved
        self.assertEqual(c[key], val)

        self.assertTrue(c.save())

        cache_files = get_cache_files()
        self.assertEqual(len(cache_files), len_initial_cache_files)
        self.assertTrue(cache_files[0].startswith('relstorage-cache2-'), cache_files)

        # Loading it works
        c2 = self._makeOne(cache_local_dir=temp_dir)
        c2.restore()
        self.assertEqual(c2[key], val)
        cache_files = get_cache_files()
        self.assertEqual(len_initial_cache_files, len(cache_files))

        # Add a new key and saving updates the existing file.
        key2 = (1, 1)
        val2 = (b'def', 1)
        # Non matching keys aren't allowed
        with self.assertRaises(AssertionError):
            c2[(1, 0)] = val2
        c2[key2] = val2
        c2.__getitem__(key2) # increment

        c2.save()
        new_cache_files = get_cache_files()
        # Same file still
        self.assertEqual(cache_files, new_cache_files)

        # And again
        cache_files = new_cache_files
        c2.save()
        new_cache_files = get_cache_files()
        self.assertEqual(cache_files, new_cache_files)

        # Notice, though, that we normalized the tid value
        # on reading.
        c3 = self._makeOne(cache_local_dir=temp_dir)
        c3.restore()
        self.assertEqual(c3[key], val)
        self.assertEqual(c3[key2], val2)

        # If we corrupt the file, it is silently ignored and removed
        for f in new_cache_files:
            with open(os.path.join(temp_dir, f), 'wb') as f:
                f.write(b'Nope!')

        c3 = self._makeOne(cache_local_dir=temp_dir)
        c3.restore()
        self.assertEqual(c3[key], None)
        cache_files = get_cache_files()
        self.assertEqual(len_initial_cache_files, len(cache_files))

        c3.remove_invalid_persistent_oids([0])

        # At no point did we spawn extra threads
        self.assertEqual(1, threading.active_count())



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A foundation for history-free RelStorage tests"""
from __future__ import absolute_import
from __future__ import print_function

import time

from ZODB.Connection import TransactionMetaData
from ZODB.serialize import referencesf
from ZODB.POSException import POSKeyError
from ZODB.tests.ConflictResolution import PCounter
from ZODB.tests.PackableStorage import ZERO
from ZODB.tests.PackableStorage import Root
from ZODB.tests.PackableStorage import dumps
from ZODB.tests.PackableStorage import pdumps
from ZODB.tests.StorageTestBase import zodb_pickle
from ZODB.tests.StorageTestBase import zodb_unpickle
# This class is sadly not cooperative with its superclass,
# so we need to explicitly place it at the back of the MRO.
from ZODB.tests.util import TestCase as ZODBTestCase

from relstorage._util import bytes8_to_int64

from relstorage.tests.RecoveryStorage import BasicRecoveryStorage
from relstorage.tests.RecoveryStorage import UndoableRecoveryStorage
from relstorage.tests.reltestbase import GenericRelStorageTests
from relstorage.tests.reltestbase import AbstractFromFileStorage
from relstorage.tests.reltestbase import AbstractToFileStorage


class HistoryFreeRelStorageTests(GenericRelStorageTests, ZODBTestCase):
    # pylint:disable=too-many-ancestors,abstract-method,too-many-locals,too-many-statements

    keep_history = False

    # This overrides certain tests so they work with a storage that
    # collects garbage but does not retain old versions.

    def _dostore(self, *args, **kwargs): # pylint:disable=arguments-differ,signature-differs
        result = super(HistoryFreeRelStorageTests, self)._dostore(*args, **kwargs)
        # Finish the transaction and update our view of the database.
        self._storage.afterCompletion()
        self._storage.poll_invalidations()
        return result

    def checkPackAllRevisions(self):
        from relstorage._compat import loads
        self._initroot()
        eq = self.assertEqual
        raises = self.assertRaises
        # Create a `persistent' object
        obj = self._newobj()
        oid = obj.getoid()
        obj.value = 1
        storage = self._storage
        # Commit three different revisions
        tid1 = self._dostoreNP(oid, data=pdumps(obj))
        storage.poll_invalidations()
        obj.value = 2
        tid2 = self._dostoreNP(oid, revid=tid1, data=pdumps(obj))
        storage.poll_invalidations()

        obj.value = 3
        tid3 = self._dostoreNP(oid, revid=tid2, data=pdumps(obj))
        storage.poll_invalidations()

        # Now make sure only the latest revision can be extracted
        for tid in tid1, tid2:
            __traceback_info__ = tid, tid1, tid2
            with raises(KeyError):
                storage.loadSerial(oid, tid)

        data = storage.loadSerial(oid, tid3)
        pobj = loads(bytes(data))
        eq(pobj.getoid(), oid)
        eq(pobj.value, 3)

        storage.pack(self._storage.lastTransactionInt() + 1, referencesf)

        # All revisions of the object should be gone, since there is no
        # reference from the root object to this object.
        for tid in tid1, tid2, tid3:
            __traceback_info__ = oid, tid
            with raises(KeyError):
                storage.loadSerial(oid, tid)
        with raises(KeyError):
            storage.load(oid)

    def checkPackJustOldRevisions(self):
        eq = self.assertEqual
        raises = self.assertRaises
        loads = self._makeloader()
        # Create a root object.  This can't be an instance of Object,
        # otherwise the pickling machinery will serialize it as a persistent
        # id and not as an object that contains references (persistent ids) to
        # other objects.
        root = Root()
        # Create a persistent object, with some initial state
        obj = self._newobj()
        oid = obj.getoid()
        # Link the root object to the persistent object, in order to keep the
        # persistent object alive.  Store the root object.
        root.obj = obj
        root.value = 0
        revid0 = self._dostoreNP(ZERO, data=dumps(root))
        # Make sure the root can be retrieved
        data, revid = self._storage.load(ZERO, '')
        eq(revid, revid0)
        eq(loads(data).value, 0)
        # Commit three different revisions of the other object
        obj.value = 1
        revid1 = self._dostoreNP(oid, data=pdumps(obj))
        obj.value = 2
        revid2 = self._dostoreNP(oid, revid=revid1, data=pdumps(obj))
        obj.value = 3
        revid3 = self._dostoreNP(oid, revid=revid2, data=pdumps(obj))
        # Now make sure only the latest revision can be extracted
        __traceback_info__ = [bytes8_to_int64(x) for x in (oid, revid1, revid2)]
        raises(KeyError, self._storage.loadSerial, oid, revid1)
        raises(KeyError, self._storage.loadSerial, oid, revid2)
        data = self._storage.loadSerial(oid, revid3)
        pobj = loads(data)
        eq(pobj.getoid(), oid)
        eq(pobj.value, 3)
        # Now pack.  The object should stay alive because it's pointed
        # to by the root.
        self._storage.pack(self._storage.lastTransactionInt(), referencesf)
        # Make sure the revisions are gone, but that object zero and revision
        # 3 are still there and correct
        data, revid = self._storage.load(ZERO, '')
        eq(revid, revid0)
        eq(loads(data).value, 0)
        raises(KeyError, self._storage.loadSerial, oid, revid1)
        raises(KeyError, self._storage.loadSerial, oid, revid2)
        data = self._storage.loadSerial(oid, revid3)
        pobj = loads(data)
        eq(pobj.getoid(), oid)
        eq(pobj.value, 3)
        data, revid = self._storage.load(oid, '')
        eq(revid, revid3)
        pobj = loads(data)
        eq(pobj.getoid(), oid)
        eq(pobj.value, 3)

    def checkPackOnlyOneObject(self):
        eq = self.assertEqual
        raises = self.assertRaises
        loads = self._makeloader()
        # Create a root object.  This can't be an instance of Object,
        # otherwise the pickling machinery will serialize it as a persistent
        # id and not as an object that contains references (persistent ids) to
        # other objects.
        root = Root()
        # Create a persistent object, with some initial state
        obj1 = self._newobj()
        oid1 = obj1.getoid()
        # Create another persistent object, with some initial state.
        obj2 = self._newobj()
        oid2 = obj2.getoid()
        # Link the root object to the persistent objects, in order to keep
        # them alive.  Store the root object.
        root.obj1 = obj1
        root.obj2 = obj2
        root.value = 0
        revid0 = self._dostoreNP(ZERO, data=dumps(root))
        # Make sure the root can be retrieved
        data, revid = self._storage.load(ZERO, '')
        eq(revid, revid0)
        eq(loads(data).value, 0)
        # Commit three different revisions of the first object
        obj1.value = 1
        revid1 = self._dostoreNP(oid1, data=pdumps(obj1))
        obj1.value = 2
        revid2 = self._dostoreNP(oid1, revid=revid1, data=pdumps(obj1))
        obj1.value = 3
        revid3 = self._dostoreNP(oid1, revid=revid2, data=pdumps(obj1))
        # Now make sure only the latest revision can be extracted
        raises(KeyError, self._storage.loadSerial, oid1, revid1)
        raises(KeyError, self._storage.loadSerial, oid1, revid2)
        data = self._storage.loadSerial(oid1, revid3)
        pobj = loads(data)
        eq(pobj.getoid(), oid1)
        eq(pobj.value, 3)
        # Now commit a revision of the second object
        obj2.value = 11
        revid4 = self._dostoreNP(oid2, data=pdumps(obj2))
        # And make sure the revision can be extracted
        data = self._storage.loadSerial(oid2, revid4)
        pobj = loads(data)
        eq(pobj.getoid(), oid2)
        eq(pobj.value, 11)
        # Now pack just revisions 1 and 2 of object1.  Object1's current
        # revision should stay alive because it's pointed to by the root, as
        # should Object2's current revision.
        now = packtime = time.time()
        while packtime <= now:
            packtime = time.time()
        self._storage.pack(packtime, referencesf)
        # Make sure the revisions are gone, but that object zero, object2, and
        # revision 3 of object1 are still there and correct.
        data, revid = self._storage.load(ZERO, '')
        eq(revid, revid0)
        eq(loads(data).value, 0)
        raises(KeyError, self._storage.loadSerial, oid1, revid1)
        raises(KeyError, self._storage.loadSerial, oid1, revid2)
        data = self._storage.loadSerial(oid1, revid3)
        pobj = loads(data)
        eq(pobj.getoid(), oid1)
        eq(pobj.value, 3)
        data, revid = self._storage.load(oid1, '')
        eq(revid, revid3)
        pobj = loads(data)
        eq(pobj.getoid(), oid1)
        eq(pobj.value, 3)
        data, revid = self._storage.load(oid2, '')
        eq(revid, revid4)
        eq(loads(data).value, 11)
        data = self._storage.loadSerial(oid2, revid4)
        pobj = loads(data)
        eq(pobj.getoid(), oid2)
        eq(pobj.value, 11)

    def checkRSResolve(self):
        # ZODB.tests.ConflictResolution.ConflictResolvingStorage has a checkResolve
        # with a different signature (as of 4.4.0) that we were unintentionally(?)
        # shadowing, hence the weird name.
        obj = PCounter()
        obj.inc()

        oid = self._storage.new_oid()

        revid1 = self._dostoreNP(oid, data=zodb_pickle(obj))

        obj.inc()
        obj.inc()

        # The effect of committing two transactions with the same
        # pickle is to commit two different transactions relative to
        # revid1 that add two to _value.

        # open s1 at this point of time.
        s1 = self._storage.new_instance()
        # start a load transaction in s1
        s1.poll_invalidations()

        # commit a change not visible to s1
        _revid2 = self._dostoreNP(oid, revid=revid1, data=zodb_pickle(obj))

        # commit a conflicting change using s1
        main_storage = self._storage
        self._storage = s1
        try:
            # we can resolve this conflict because s1 has an open
            # transaction that can read the old state of the object.
            _revid3 = self._dostoreNP(oid, revid=revid1, data=zodb_pickle(obj))
            s1.release()
        finally:
            self._storage = main_storage

        # If we don't restart our load connection,
        # we will still read the old state.
        data, _serialno = self._storage.load(oid, '')
        inst = zodb_unpickle(data)
        self.assertEqual(inst._value, 3)

        self._storage.poll_invalidations()
        data, _serialno = self._storage.load(oid, '')
        inst = zodb_unpickle(data)
        self.assertEqual(inst._value, 5)


    def checkImplementsIExternalGC(self):
        from zope.interface.verify import verifyObject
        import ZODB.interfaces
        verifyObject(ZODB.interfaces.IExternalGC, self._storage)

        # Now do it.
        from ZODB.utils import z64
        db = self._closing(ZODB.DB(self._storage)) # create the root
        conn = self._closing(db.open())
        storage = conn._storage
        _state, tid = storage.load(z64)

        # copied from zc.zodbdgc
        t = TransactionMetaData()
        storage.tpc_begin(t)
        count = storage.deleteObject(z64, tid, t)
        self.assertEqual(count, 1)
        # Doing it again will do nothing because it's already
        # gone.
        count = storage.deleteObject(z64, tid, t)
        self.assertEqual(count, 0)

        storage.tpc_vote(t)
        storage.tpc_finish(t)


        with self.assertRaises(POSKeyError):
            storage.load(z64)

        conn.close()
        db.close()

    def checkIteratorIsConsistent(self):
        # History free iterators should be consistent
        # until a sync() is done.
        # https://github.com/zodb/relstorage/issues/344
        root = Root()
        self._dostoreNP(ZERO, data=dumps(root))

        storage = self._storage
        it = self._storage.iterator()
        self.assertEqual(1, len(it))
        it.close()

        obj1 = self._newobj()
        oid1 = obj1.getoid()
        storage2 = self._storage.new_instance()
        self._storage = storage2
        self._dostoreNP(oid1, data=dumps(obj1))
        self._storage = storage

        it2 = storage2.iterator()
        self.assertEqual(2, len(it2))
        it2.close()

        it = self._storage.iterator()
        self.assertEqual(1, len(it))
        it.close()

        self._storage.sync()
        it = self._storage.iterator()
        self.assertEqual(2, len(it))
        it.close()

        storage2.close()

    def checkSetMinOid(self):
        # Verify that OID allocation goes as expected when we use
        # set_min_oid. There have been problems in the past translating
        # between user-space OIDs and the 16-range gapped OIDs that are actually
        # stored.

        # TODO: This test is independent of history keeping status.
        # We should have file for all those type tests and only mix it in
        # to one of the test classes.

        storage = self._storage
        I = bytes8_to_int64
        for offset in range(1, 50):
            # Ensure no gaps.
            oid = storage.new_oid()
            self.assertEqual(offset, I(oid))

        with storage._store_connection_pool.borrowing() as store_conn:
            storage._oids.set_min_oid(store_conn, 32768)
            store_conn.commit()
        oid = storage.new_oid()
        self.assertEqual(32769, I(oid))

        # Close to the 64-bit boundary.
        with storage._store_connection_pool.borrowing() as store_conn:
            storage._oids.set_min_oid(store_conn, 2 ** 62)
            store_conn.commit()
        # Iterate through several ranges. This has been a problem in the past.
        for offset in range(1, 50):
            self.assertEqual(2 ** 62 + offset, I(storage.new_oid()))



class HistoryFreeToFileStorage(AbstractToFileStorage,
                               BasicRecoveryStorage,
                               ZODBTestCase):
    # pylint:disable=abstract-method,too-many-ancestors
    keep_history = False


class HistoryFreeFromFileStorage(AbstractFromFileStorage,
                                 UndoableRecoveryStorage,
                                 ZODBTestCase):
    # pylint:disable=abstract-method,too-many-ancestors
    keep_history = False

    def compare(self, src, dest):
        # The dest storage has a truncated copy of dest, so
        # use compare_truncated() instead of compare_exact().
        self.compare_truncated(src, dest)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from ...sql.tests import test_sql
from ..drivers import MySQLDialect

class TestMySQLDialect(test_sql.TestUpsert):
    keep_history = False
    dialect = MySQLDialect()

    insert_or_replace = (
        'INSERT INTO object_state(zoid, state, tid, state_size) '
        'VALUES (%s, %s, %s, %s) '
        'ON DUPLICATE KEY UPDATE '
        'state = VALUES(state), tid = VALUES(tid), '
        'state_size = VALUES(state_size)'
    )

    insert_or_replace_subquery = (
        'INSERT INTO object_state(zoid, tid, state, state_size) '
        'SELECT zoid, %s, state, COALESCE(LENGTH(state), 0) FROM temp_store '
        'ORDER BY zoid '
        'ON DUPLICATE KEY UPDATE '
        'state = VALUES(state), tid = VALUES(tid), '
        'state_size = VALUES(state_size)'
    )

    upsert_unconstrained_subquery = (
        'INSERT INTO object_state(zoid, tid, state, state_size) '
        'SELECT zoid, %s, state, COALESCE(LENGTH(state), 0) FROM temp_store '
        'ON DUPLICATE KEY UPDATE state = VALUES(state), '
        'tid = VALUES(tid), state_size = VALUES(state_size)'
    )



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from relstorage.tests import TestCase
from relstorage.tests import MockOptions
from relstorage.tests import MockDriver

from .. import mover

class PGMockDriver(MockDriver):
    supports_copy = False

@unittest.skip("Needs moved to test__sql")
class TestFunctions(TestCase):
    # pylint:disable=no-member
    def _prepare1(self, query, name='prepped', datatypes=()):
        return mover.to_prepared_queries(name, [query], datatypes)[0]

    def test_prepared_no_param_no_datatype(self):
        q = 'SELECT foo FROM bar'
        p = self._prepare1(q)
        self.assertEqual(
            'PREPARE prepped AS ' + q,
            p
        )

    def test_prepared_one_param_no_datatype(self):
        q = 'SELECT foo FROM bar WHERE foo = %s'
        p = self._prepare1(q)
        self.assertEqual(
            'PREPARE prepped AS SELECT foo FROM bar WHERE foo = $1',
            p
        )

    def test_prepared_one_param_one_datatype(self):
        q = 'SELECT foo FROM bar WHERE foo = %s'
        p = self._prepare1(q, datatypes=['int'])
        self.assertEqual(
            'PREPARE prepped (int) AS SELECT foo FROM bar WHERE foo = $1',
            p
        )

    def test_prepared_two_param_two_datatype(self):
        q = 'SELECT foo FROM bar WHERE foo = %s and biz = %s'
        p = self._prepare1(q, datatypes=['int', 'bigint'])
        self.assertEqual(
            'PREPARE prepped (int, bigint) AS SELECT foo FROM bar WHERE foo = $1 '
            'and biz = $2',
            p
        )

    maxDiff = None

    def test_prepare_load_current(self):
        self.assertEqual(
            mover.PostgreSQLObjectMover._prepare_load_current_queries,
            [
                'PREPARE load_current (BIGINT) AS SELECT state, tid\n'
                '        FROM current_object\n'
                '        JOIN object_state USING(zoid, tid)\n'
                '        WHERE zoid = $1',
                'PREPARE load_current (BIGINT) AS SELECT state, tid\n'
                '        FROM object_state\n'
                '        WHERE zoid = $1'
            ]
        )

class TestPostgreSQLObjectMover(TestCase):

    def _getClass(self):
        return mover.PostgreSQLObjectMover

    def _makeOne(self, **options):
        return self._getClass()(PGMockDriver(),
                                MockOptions.from_args(**options))

    _expected_move_from_temp_hf_insert_query = 'EXECUTE move_from_temp(%s)'

    def test_prep_statements_hf(self):
        inst = self._makeOne(keep_history=False)
        self.assertTrue(
            str(inst._move_from_temp_hf_upsert_query).startswith(
                'EXECUTE rs_prep_stmt'
            )
        )



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A memcache-like module sufficient for testing without an actual memcache.
"""

data = {}

class Client(object):

    def __init__(self, servers):
        self.servers = servers

    def delete(self, key):
        data.pop(key, None)

    def get(self, key):
        return data.get(key)

    def get_multi(self, keys):
        return dict((key, data.get(key)) for key in keys)

    def set(self, key, value):
        data[key] = value

    def set_multi(self, d):
        data.update(d)

    def add(self, key, value):
        if key not in data:
            data[key] = value

    def incr(self, key):
        value = data.get(key)
        if value is None:
            return None
        value = int(value) + 1
        data[key] = value
        return value

    def flush_all(self, **_kwargs):
        data.clear()

    def disconnect_all(self):
        # no-op
        pass
